# Test for one implementation of the interface
import re
from unittest import TestCase

from lexicon.tests.providers import integration_tests


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class MythicBeastsProviderTests(TestCase, integration_tests.IntegrationTestsV2):
    """Integration tests for Mythic Beasts provider"""

    provider_name = "mythicbeasts"
    domain = "lexitus.co.uk"

    def _filter_post_data_parameters(self):
        return ["access_token"]

    def _filter_headers(self):
        return ["Authorization"]

    def _filter_response(self, response):
        """See `IntegrationTests._filter_response` for more information on how
        to filter the provider response."""

        if "string" in response["body"]:
            response["body"]["string"] = re.sub(
                br"\"access_token\":\"[\w-]+\"",
                b'"access_token": "DUMMY_TOKEN"',
                response["body"]["string"],
            )

        return response

    def _test_fallback_fn(self):
        return lambda x: "placeholder_" + x if x not in ("auth_token") else ""



"""Integration tests for InternetBS"""
import os
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class InternetbsProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for InternetBS"""

    provider_name = "internetbs"
    domain = "world-of-ysera.com"

    def _filter_query_parameters(self):
        return ["ApiKey", "Password"]

    def _test_parameters_overrides(self):
        # workaround ENV problems during testing
        env_key = os.environ.get("LEXICON_INTERNETBS_AUTH_KEY")
        env_password = os.environ.get("LEXICON_INTERNETBS_AUTH_PASSWORD")
        return {"auth_key": env_key, "auth_password": env_password}



"""Integration tests for INWX"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class InwxProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for INWX"""

    provider_name = "inwx"
    domain = "lexicon-testrunner.com"

    def _test_parameters_overrides(self):
        return {"endpoint": "https://api.ote.domrobot.com/xmlrpc/"}



"""Integration tests for EasyName"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class EasynameProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for EasyName"""

    provider_name = "easyname"
    domain = "fivekindfilms.com"

    def _filter_post_data_parameters(self):
        return ["emailAddress", "password"]

    def _filter_headers(self):
        return ["Cookie"]

    def _test_fallback_fn(self):
        return lambda x: "placeholder_" + x if x != "priority" else ""



"""Main module of Lexicon. Defines the Client class, that holds all Lexicon logic."""
import importlib
import logging
import os
from typing import Dict, List, Optional, Type, Union

import tldextract  # type: ignore

from lexicon import config as helper_config
from lexicon import discovery
from lexicon.exceptions import ProviderNotAvailableError
from lexicon.providers.base import Provider


class Client(object):
    """This is the Lexicon client, that will execute all the logic."""

    def __init__(
        self, config: Optional[Union[helper_config.ConfigResolver, Dict]] = None
    ):
        if not config:
            # If there is not config specified, we load a non-interactive configuration.
            self.config = helper_config.non_interactive_config_resolver()
        elif not isinstance(config, helper_config.ConfigResolver):
            # If config is not a ConfigResolver, we are in a legacy situation.
            # We protect this part of the Client API.
            self.config = helper_config.legacy_config_resolver(config)
        else:
            self.config = config

        # Validate configuration
        self._validate_config()

        runtime_config = {}

        # Process domain, strip subdomain
        try:
            domain_extractor = tldextract.TLDExtract(
                cache_dir=_get_tldextract_cache_path(), include_psl_private_domains=True
            )
        except TypeError:
            domain_extractor = tldextract.TLDExtract(
                cache_file=_get_tldextract_cache_path(), include_psl_private_domains=True  # type: ignore
            )
        domain_parts = domain_extractor(self.config.resolve("lexicon:domain"))
        runtime_config["domain"] = f"{domain_parts.domain}.{domain_parts.suffix}"

        delegated = self.config.resolve("lexicon:delegated")
        if delegated:
            # handle delegated domain
            delegated = str(delegated).rstrip(".")
            initial_domain = str(runtime_config.get("domain"))
            if delegated != initial_domain:
                # convert to relative name
                if delegated.endswith(initial_domain):
                    delegated = delegated[: -len(initial_domain)]
                    delegated = delegated.rstrip(".")
                # update domain
                runtime_config["domain"] = f"{delegated}.{initial_domain}"

        self.action = self.config.resolve("lexicon:action")
        self.provider_name = self.config.resolve(
            "lexicon:provider_name"
        ) or self.config.resolve("lexicon:provider")

        if not self.provider_name:
            raise ValueError("Could not resolve provider name.")

        self.config.add_config_source(helper_config.DictConfigSource(runtime_config), 0)

        provider_module = importlib.import_module(
            "lexicon.providers." + self.provider_name
        )
        provider_class: Type[Provider] = getattr(provider_module, "Provider")
        self.provider = provider_class(self.config)

    def execute(self) -> Union[bool, List[Dict]]:
        """Execute provided configuration in class constructor to the DNS records"""
        self.provider.authenticate()
        identifier = self.config.resolve("lexicon:identifier")
        record_type = self.config.resolve("lexicon:type")
        name = self.config.resolve("lexicon:name")
        content = self.config.resolve("lexicon:content")

        if self.action == "create":
            if not record_type or not name or not content:
                raise ValueError("Missing record_type, name or content parameters.")
            return self.provider.create_record(record_type, name, content)

        if self.action == "list":
            return self.provider.list_records(record_type, name, content)

        if self.action == "update":
            return self.provider.update_record(identifier, record_type, name, content)

        if self.action == "delete":
            return self.provider.delete_record(identifier, record_type, name, content)

        raise ValueError(f"Invalid action statement: {self.action}")

    def _validate_config(self) -> None:
        provider_name = self.config.resolve("lexicon:provider_name")
        if not provider_name:
            raise AttributeError("provider_name")

        try:
            available = discovery.find_providers()[provider_name]
        except KeyError:
            raise ProviderNotAvailableError(
                f"This provider ({provider_name}) is not supported by Lexicon."
            )
        else:
            if not available:
                raise ProviderNotAvailableError(
                    f"This provider ({provider_name}) has required dependencies that are missing. "
                    f"Please install lexicon[{provider_name}] first."
                )

        if not self.config.resolve("lexicon:action"):
            raise AttributeError("action")
        if not self.config.resolve("lexicon:domain"):
            raise AttributeError("domain")
        if not self.config.resolve("lexicon:type"):
            raise AttributeError("type")


def _get_tldextract_cache_path() -> str:
    if os.environ.get("TLDEXTRACT_CACHE_FILE"):
        logging.warning(
            "TLD_EXTRACT_CACHE_FILE environment variable is deprecated, please use TLDEXTRACT_CACHE_PATH instead."
        )
        os.environ["TLDEXTRACT_CACHE_PATH"] = os.environ["TLDEXTRACT_CACHE_FILE"]

    return os.path.expanduser(
        os.environ.get("TLDEXTRACT_CACHE_PATH", os.path.join("~", ".lexicon_tld_set"))
    )



"""
A note about running these tests against a Infoblox Environment

1. Make sure the NIOS Version support WAPI 2.6.1
2. Have a valid Certificate from a public CA installed at the Infoblox
3. Create a Authoritative zone test.local in a view (default if no views are created)
4. Create a User with permissions RW permissions
   for the zone test.local and enable the User for API

Environment Variables work fine when envoking lexicon manually
LEXICON_INFOBLOX_AUTH_USER={user} LEXICON_INFOBLOX_AUTH_PSW={password}
    lexicon infoblox --ib-host dns1.int.metro-cc.com --ib-view internal
        create test.local A --content 10.10.10.11 --name lexicon1
Invoking the py.test however fails
LEXICON_LIVE_TESTS=true LEXICON_INFOBLOX_AUTH_USER={username} LEXICON_INFOBLOX_AUTH_PSW={password}
    py.test tests/providers/test_infoblox.py
Both parameters are populated with:
auth_user = placeholder_auth_user
auth_psw = placeholder_auth_psw
"""
import os
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class InfobloxProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Infoblox"""

    provider_name = "infoblox"
    domain = "test.local"

    def _test_parameters_overrides(self):
        # Workaround ENV problems during testing
        env_user = os.environ.get("LEXICON_INFOBLOX_AUTH_USER", "infoblox")
        env_psw = os.environ.get("LEXICON_INFOBLOX_AUTH_PSW", "default")
        return {
            "ib_host": "dns1.int.metro-cc.com",
            "ib_view": "internal",
            "auth_user": env_user,
            "auth_psw": env_psw,
        }

    def _filter_headers(self):
        return ["Authorization", "Cookie", "set-cookie"]



"""
Implement the Azure DNS provider.
The Azure DNS API is well documented, with meaningful examples.
The major difficulties are the authentication, that uses OAuth, and require several authentication
parameters, and the fact that records are manipulated as recordsets (entries with multiple values
like TXT are hold on a unique TXTRecords object), and so require methods to go from the Azure
representation of a DNS record to the Lexicon one. To simplify the implementation then, update
actions are managed using the create and delete actions: updating a record consists in removing the
old record, then creating a new record with updated parameters. Note also that Azure DNS does not
support id on records (as they are hold in a recordset) so an id is generated on the fly.
"""
import binascii
import logging

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

AZURE_AD_URL = "https://login.microsoftonline.com"
MANAGEMENT_URL = "https://management.azure.com"
API_VERSION = "2018-03-01-preview"
NAMESERVER_DOMAINS = [
    "azure-dns.com",
    "azure-dns.net",
    "azure-dns.org",
    "azure-dns.info",
]

SUPPORTED_RECORDS = {"A", "AAAA", "CNAME", "MX", "NS", "SOA", "TXT", "SRV"}


def provider_parser(subparser):
    """Generate a subparser for Azure DNS"""
    subparser.description = """
        The Azure provider orchestrates the DNS zones hosted in a resource group for a subscription
        in Microsoft Azure Cloud. To authenticate, an App registration must be created in an Azure
        Active Directory. This App registration must be granted Admin for API permissions to
        Domain.ReadWrite.All" to this Active Directory, and must have a usable Client secret.
    """
    subparser.add_argument(
        "--auth-client-id",
        help="specify the client ID (aka application ID) " "of the App registration",
    )
    subparser.add_argument(
        "--auth-client-secret",
        help="specify the client secret of the App " "registration",
    )
    subparser.add_argument(
        "--auth-tenant-id",
        help="specify the tenant ID (aka directory ID) of " "the App registration",
    )
    subparser.add_argument(
        "--auth-subscription-id",
        help="specify the subscription ID attached " "to the resource group",
    )
    subparser.add_argument(
        "--resource-group",
        help="specify the resource group hosting the DNS " "zone to edit",
    )


class Provider(BaseProvider):
    """Provider for Azure DNS"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._access_token = None

    def _list_records(self, rtype=None, name=None, content=None):
        result = self._get(f"/{(rtype if rtype else 'recordsets')}")

        records = []
        for raw_record in result["value"]:
            rtype = raw_record["type"].replace("Microsoft.Network/dnszones/", "")
            if rtype not in SUPPORTED_RECORDS:
                continue
            for value in _get_values_from_recordset(rtype, raw_record):
                record = {
                    "type": rtype,
                    "name": self._full_name(raw_record["name"]),
                    "ttl": raw_record["properties"]["TTL"],
                    "content": value,
                }
                record["id"] = _identifier(record)
                records.append(record)

        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def _create_record(self, rtype, name, content):
        identifier = self._create_record_internal(rtype, name, content)

        LOGGER.debug("create_record: %s", identifier)

        return True

    def _create_record_internal(self, rtype, name, content):
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        identifier = _identifier(
            {"type": rtype, "name": self._full_name(name), "content": content}
        )

        records = self._list_records(rtype, name)

        if [record for record in records if record["id"] == identifier]:
            LOGGER.debug("create_record (ignored, duplicate): %s", identifier)
            return True

        values = [record["content"] for record in records]
        values.append(content)

        properties = _build_recordset_from_values(rtype, values)

        ttl = self._get_lexicon_option("ttl")
        if ttl:
            properties["TTL"] = ttl

        self._put(
            f"/{rtype}/{self._relative_name(name)}", data={"properties": properties}
        )

        return identifier

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier and (not rtype or not name):
            raise Exception("Error, identifier or rtype+name parameters are required.")

        if identifier:
            records = self._list_records()
            records_to_update = [
                record for record in records if record["id"] == identifier
            ]
        else:
            records_to_update = self._list_records(rtype=rtype, name=name)

        if not records_to_update:
            raise Exception(
                f"Error, could not find a record for given identifier: {identifier}"
            )

        if len(records_to_update) > 1:
            LOGGER.warning(
                "Warning, multiple records found for given parameters, "
                "only first one will be updated: %s",
                records_to_update,
            )

        identifier = identifier if identifier else records_to_update[0]["id"]
        rtype = rtype if rtype else records_to_update[0]["type"]
        name = name if name else records_to_update[0]["name"]
        content = content if content else records_to_update[0]["content"]

        self._delete_record_internal(identifier=identifier)
        self._create_record_internal(rtype=rtype, name=name, content=content)

        LOGGER.debug(
            "update_record: %s => %s",
            identifier,
            _identifier({"type": rtype, "name": name, "content": content}),
        )

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        self._delete_record_internal(identifier, rtype, name, content)

        LOGGER.debug("delete_records: %s %s %s %s", identifier, rtype, name, content)

        return True

    def _delete_record_internal(
        self, identifier=None, rtype=None, name=None, content=None
    ):
        result = self._get(f"/{(rtype if rtype else 'recordsets')}")

        to_delete = []
        to_shrink = []
        for record in result["value"]:
            record_rtype = record["type"].replace("Microsoft.Network/dnszones/", "")
            if record_rtype not in SUPPORTED_RECORDS:
                continue
            new_values = []
            values = _get_values_from_recordset(record_rtype, record)
            for value in values:
                if identifier is not None and identifier != _identifier(
                    {
                        "type": record_rtype,
                        "name": self._full_name(record["name"]),
                        "content": value,
                    }
                ):
                    new_values.append(value)
                elif identifier is None:
                    matching_rtype = rtype is None or record_rtype == rtype
                    matching_name = name is None or self._full_name(
                        name
                    ) == self._full_name(record["name"])
                    matching_content = content is None or value == content
                    if not (matching_rtype and matching_name and matching_content):
                        new_values.append(value)

            to_modify = len(values) != len(new_values)
            if to_modify:
                record["properties"].update(
                    _build_recordset_from_values(record_rtype, new_values)
                )
                if new_values:
                    to_shrink.append(record)
                else:
                    to_delete.append(record)

        for record in to_delete:
            record_rtype = record["type"].replace("Microsoft.Network/dnszones/", "")
            self._delete(f"/{record_rtype}/{self._relative_name(record['name'])}")
        for record in to_shrink:
            record_rtype = record["type"].replace("Microsoft.Network/dnszones/", "")
            self._request(
                "PATCH",
                f"/{record_rtype}/{self._relative_name(record['name'])}",
                data={"properties": record["properties"]},
            )

    def _authenticate(self):
        tenant_id = self._get_provider_option("auth_tenant_id")
        client_id = self._get_provider_option("auth_client_id")
        client_secret = self._get_provider_option("auth_client_secret")
        subscription_id = self._get_provider_option("auth_subscription_id")
        resource_group = self._get_provider_option("resource_group")

        assert tenant_id
        assert client_id
        assert client_secret
        assert subscription_id
        assert resource_group

        url = f"{AZURE_AD_URL}/{tenant_id}/oauth2/token"
        data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "resource": MANAGEMENT_URL,
        }

        result = requests.post(url, data=data)
        result.raise_for_status()

        self._access_token = result.json()["access_token"]

        url = "{0}/subscriptions/{1}/resourceGroups/{2}/providers/Microsoft.Network/dnsZones".format(
            MANAGEMENT_URL, subscription_id, resource_group
        )
        headers = {"Authorization": f"Bearer {self._access_token}"}
        params = {"api-version": API_VERSION}

        result = requests.get(url, headers=headers, params=params)
        result.raise_for_status()

        data = result.json()

        our_data = [
            one_data for one_data in data["value"] if one_data["name"] == self.domain
        ]

        if not our_data:
            raise AuthenticationError(
                "Resource group `{0}` in subscription `{1}` "
                "does not contain the DNS zone `{2}`".format(
                    resource_group, subscription_id, self.domain
                )
            )

        self.domain_id = our_data[0]["id"]

    def _request(self, action="GET", url="/", data=None, query_params=None):
        query_params = {} if not query_params else query_params.copy()
        query_params["api-version"] = API_VERSION
        headers = {"Authorization": f"Bearer {self._access_token}"}
        request = requests.request(
            action,
            MANAGEMENT_URL + self.domain_id + url,
            params=query_params,
            headers=headers,
            json=None if not data else data,
        )
        request.raise_for_status()
        if request.content:
            return request.json()
        return None


def _get_values_from_recordset(rtype, record):
    properties = record["properties"]
    values = None
    if rtype == "A":
        values = [entry["ipv4Address"] for entry in properties["ARecords"]]
    if rtype == "AAAA":
        values = [entry["ipv6Address"] for entry in properties["AAAARecords"]]
    if rtype == "CNAME":
        values = [properties["CNAMERecord"]["cname"]]
    if rtype == "MX":
        values = [entry["exchange"] for entry in properties["MXRecords"]]
    if rtype == "NS":
        values = [entry["nsdname"] for entry in properties["NSRecords"]]
    if rtype == "SOA":
        values = [properties["SOARecord"]["email"]]
    if rtype == "TXT":
        values = [
            value for entry in properties["TXTRecords"] for value in entry["value"]
        ]
    if rtype == "SRV":
        values = [entry["target"] for entry in properties["SRVRecords"]]

    if values:
        return values

    raise Exception(f"Error, `{rtype}` entries are not supported by this provider.")


def _build_recordset_from_values(rtype, values):
    recordset = None
    if rtype == "A":
        recordset = {"ARecords": [{"ipv4Address": value} for value in values]}
    if rtype == "AAAA":
        recordset = {"AAAARecords": [{"ipv6Address": value} for value in values]}
    if rtype == "CNAME":
        recordset = {"CNAMERecord": {"cname": values[0]} if values else {}}
    if rtype == "MX":
        recordset = {"MXRecords": [{"exchange": value} for value in values]}
    if rtype == "NS":
        recordset = {"NSRecords": [{"nsdname": value} for value in values]}
    if rtype == "SOA":
        recordset = {"SOARecord": {"email": values[0]} if values else {}}
    if rtype == "TXT":
        recordset = {"TXTRecords": [{"value": values}]}
    if rtype == "SRV":
        recordset = {"SRVRecords": [{"target": value} for value in values]}

    if recordset:
        return recordset

    raise Exception(f"Error, `{rtype}` entries are not supported by this provider.")


def _identifier(record):
    digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
    digest.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    digest.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    digest.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

    return binascii.hexlify(digest.finalize()).decode("utf-8")[0:7]



"""Integration tests for Conoha"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class ConohaProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Conoha"""

    provider_name = "conoha"
    domain = "narusejun.com"

    def _test_parameters_overrides(self):
        return {"region": "tyo1"}

    def _test_fallback_fn(self):
        return lambda x: None if x in ("priority") else "placeholder_" + x

    def _filter_post_data_parameters(self):
        return ["auth"]

    def _filter_headers(self):
        return ["X-Auth-Token"]



"""Integration tests for Gandi"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class GandiRPCProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Gandi on RPC"""

    provider_name = "gandi"
    domain = "reachlike.ca"
    provider_variant = "RPC"

    def _test_parameters_overrides(self):
        return {"api_protocol": "rpc"}


class GandiRESTProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Gandi on REST API"""

    provider_name = "gandi"
    domain = "t18s.fr"
    provider_variant = "REST"

    def _filter_headers(self):
        return ["Authorization"]

    def _test_parameters_overrides(self):
        return {"api_protocol": "rest"}



"""Unit tests for the Lexicon CLI parser"""
import pytest

from lexicon import parser


def test_base_provider_parser():
    baseparser = parser.generate_base_provider_parser()
    parsed = baseparser.parse_args(["list", "capsulecd.com", "TXT"])
    assert parsed.action == "list"
    assert parsed.domain == "capsulecd.com"
    assert parsed.type == "TXT"
    assert parsed.ttl is None
    assert parsed.output == "TABLE"


def test_base_provider_parser_without_domain():
    baseparser = parser.generate_base_provider_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args(["list"])


def test_base_provider_parser_without_options():
    baseparser = parser.generate_base_provider_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args([])


def test_cli_main_parser():
    baseparser = parser.generate_cli_main_parser()
    parsed = baseparser.parse_args(["cloudflare", "list", "capsulecd.com", "TXT"])
    assert parsed.provider_name == "cloudflare"
    assert parsed.action == "list"
    assert parsed.domain == "capsulecd.com"
    assert parsed.type == "TXT"
    assert parsed.output == "TABLE"


def test_cli_main_parser_without_args():
    baseparser = parser.generate_cli_main_parser()
    with pytest.raises(SystemExit):
        baseparser.parse_args([])



"""General pytest configuration for lexicon tests."""
import pytest

from lexicon import discovery


def pytest_addoption(parser):
    """Standard pytest hook invoked to add options to pytest CLI"""
    parser.addoption(
        "--xfail-providers-with-missing-deps",
        action="store_true",
        help="Skip tests on providers with optional dependencies",
    )


def pytest_runtest_setup(item):
    """Standard pytest hook invoked before each test execution"""
    try:
        skip_providers_with_optdeps = getattr(
            item.config.option, "xfail_providers_with_missing_deps"
        )
    except AttributeError:
        pass
    else:
        if skip_providers_with_optdeps:
            providers = discovery.find_providers()
            skip = [
                available
                for provider, available in providers.items()
                if provider in item.parent.name.lower()
            ]
            if skip and not skip[0]:
                pytest.xfail(
                    "Test expected to fail because --skip-providers-with-missing-deps "
                    "is set and provider has missing required dependencies."
                )



"""Unit tests for the Lexicon config mechanism"""
import pytest

from lexicon.config import ConfigResolver, ConfigSource
from lexicon.parser import generate_cli_main_parser


def test_environment_resolution(monkeypatch):
    monkeypatch.setenv("LEXICON_DELEGATED", "TEST1")
    monkeypatch.setenv("LEXICON_CLOUDFLARE_TOKEN", "TEST2")
    monkeypatch.setenv("LEXICON_CLOUDFLARE_AUTH_USERNAME", "TEST3")

    config = ConfigResolver().with_env()

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:cloudflare:auth_username") == "TEST3"
    assert config.resolve("lexicon:nonexistent") is None


def test_argparse_resolution():
    parser = generate_cli_main_parser()
    data = parser.parse_args(
        [
            "--delegated",
            "TEST1",
            "cloudflare",
            "create",
            "example.com",
            "TXT",
            "--auth-token",
            "TEST2",
        ]
    )

    config = ConfigResolver().with_args(data)

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:nonexistent") is None


def test_dict_resolution():
    dict_object = {"delegated": "TEST1", "cloudflare": {"auth_token": "TEST2"}}

    config = ConfigResolver().with_dict(dict_object)

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:nonexistent") is None


def test_config_lexicon_file_resolution(tmpdir):
    lexicon_file = tmpdir.join("lexicon.yml")
    lexicon_file.write("delegated: TEST1\ncloudflare:\n  auth_token: TEST2")

    config = ConfigResolver().with_config_file(str(lexicon_file))

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:nonexistent") is None


def test_provider_config_lexicon_file_resolution(tmpdir):
    provider_file = tmpdir.join("lexicon_cloudflare.yml")
    provider_file.write("auth_token: TEST2")

    config = ConfigResolver().with_provider_config_file(
        "cloudflare", str(provider_file)
    )

    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:nonexistent") is None


def test_provider_config_dir_resolution(tmpdir):
    lexicon_file = tmpdir.join("lexicon.yml")
    provider_file = tmpdir.join("lexicon_cloudflare.yml")
    lexicon_file.write("delegated: TEST1\ncloudflare:\n  auth_token: TEST2")
    provider_file.write("auth_username: TEST3")

    config = ConfigResolver().with_config_dir(str(tmpdir))

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:cloudflare:auth_username") == "TEST3"
    assert config.resolve("lexicon:nonexistent") is None


def test_generic_config_feeder_resolution():
    class GenericConfigSource(ConfigSource):
        def resolve(self, config_key):
            return "TEST1"

    config = ConfigResolver().with_config_source(GenericConfigSource())

    assert config.resolve("lexicon:cloudflare:auth_username") == "TEST1"
    assert config.resolve("lexicon:nonexistent") == "TEST1"


def test_legacy_dict_config_resolution():
    legacy_config = {
        "delegated": "TEST1",
        "auth_token": "TEST2",
        "provider_name": "cloudflare",
    }

    with pytest.deprecated_call():
        config = ConfigResolver().with_legacy_dict(legacy_config)

    assert config.resolve("lexicon:delegated") == "TEST1"
    assert config.resolve("lexicon:cloudflare:auth_token") == "TEST2"
    assert config.resolve("lexicon:auth_token") is None
    assert config.resolve("lexicon:nonexistent") is None


def test_prioritized_resolution(tmpdir, monkeypatch):
    lexicon_file = tmpdir.join("lexicon.yml")
    lexicon_file.write("cloudflare:\n  auth_token: TEST1")

    monkeypatch.setenv("LEXICON_CLOUDFLARE_AUTH_TOKEN", "TEST2")

    assert (
        ConfigResolver()
        .with_config_file(str(lexicon_file))
        .with_env()
        .resolve("lexicon:cloudflare:auth_token")
        == "TEST1"
    )
    assert (
        ConfigResolver()
        .with_env()
        .with_config_file(str(lexicon_file))
        .resolve("lexicon:cloudflare:auth_token")
        == "TEST2"
    )



"""Module provider for RcodeZero"""
import hashlib
import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["rcode0.net"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument("--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for RcodeZero"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._zone_data = None
        self.api_endpoint = "https://my.rcodezero.at/api/v1"

    def _authenticate(self):
        if self._zone_data is None:
            self._zone_data = self._get("/zones/" + self.domain)

        self.domain_id = self.domain

    # Create record.

    def _create_record(self, rtype, name, content):
        rname = self._fqdn_name(name)
        newcontent = self._clean_content(rtype, content)

        updated_data = {
            "name": rname,
            "type": rtype,
            "records": [],
            "ttl": self._get_lexicon_option("ttl") or 600,
            "changetype": "ADD",
        }

        updated_data["records"].append({"content": newcontent, "disabled": False})

        payload = self._get(f"/zones/{self.domain_id}/rrsets?page_size=-1")

        for rrset in payload["data"]:
            if rrset["name"] == rname and rrset["type"] == rtype:
                updated_data["ttl"] = rrset["ttl"]

                for record in rrset["records"]:
                    if record["content"] != newcontent:
                        updated_data["records"].append(
                            {
                                "content": record["content"],
                                "disabled": record["disabled"],
                            }
                        )
                    updated_data["changetype"] = "UPDATE"
                break

        request = [updated_data]
        LOGGER.debug("request: %s", request)

        self._patch("/zones/" + self.domain + "/rrsets", data=request)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_obj = {"per_page": 100}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["name"] = self._full_name(name)
        if content:
            filter_obj["content"] = content

        payload = self._get(f"/zones/{self.domain_id}/rrsets?page_size=-1")

        records = []
        for rrset in payload["data"]:
            if (
                name is None or self._fqdn_name(rrset["name"]) == self._fqdn_name(name)
            ) and (rtype is None or rrset["type"] == rtype):
                for record in rrset["records"]:
                    if content is None or record["content"] == self._clean_content(
                        rtype, content
                    ):
                        # rcode0 does not have a record id, so lets create one
                        processed_record = {
                            "type": rrset["type"],
                            "name": self._full_name(rrset["name"]),
                            "ttl": rrset["ttl"],
                            "content": self._unclean_content(
                                rrset["type"], record["content"]
                            ),
                            "id": self._make_identifier(
                                rrset["type"], rrset["name"], record["content"]
                            ),
                        }
                        records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        self._delete_record(identifier, rtype, name, None)
        return self._create_record(rtype, name, content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):

        LOGGER.debug("delete %s %s %s %s", identifier, rtype, name, content)
        if identifier is None and (rtype is None or name is None):
            raise Exception("Must specify at least id or  both rtype and name")

        payload = self._get(f"/zones/{self.domain_id}/rrsets?page_size=-1")

        if identifier is not None:
            rtype, name, content = self._parse_identifier(identifier, payload)

        update_data = None
        for rrset in payload["data"]:
            if rrset["type"] == rtype and self._fqdn_name(
                rrset["name"]
            ) == self._fqdn_name(name):
                update_data = rrset

                if content is None:
                    update_data["records"] = []
                    update_data["changetype"] = "DELETE"
                else:
                    new_record_list = []
                    for record in update_data["records"]:
                        if (
                            self._clean_content(rrset["type"], content)
                            != record["content"]
                        ):
                            new_record_list.append(record)

                    update_data["records"] = new_record_list
                    if new_record_list:
                        update_data["changetype"] = "UPDATE"
                    else:
                        update_data["changetype"] = "DELETE"
                break

        if update_data is not None:
            request = [update_data]
            LOGGER.debug("request: %s", request)
            self._patch("/zones/" + self.domain + "/rrsets", data=request)

        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Authorization": "Bearer " + self._get_provider_option("auth_token"),
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        if response.status_code >= 400:
            LOGGER.debug("Bad Request: %s", response.text)
        response.raise_for_status()
        return response.json()

    # generate a unique id for a give record
    def _make_identifier(self, rtype, name, content):
        sha256 = hashlib.sha256()
        sha256.update(("type=" + rtype + ",").encode("utf-8"))
        sha256.update(("name=" + name + ",").encode("utf-8"))
        sha256.update(("content=" + content + ",").encode("utf-8"))
        return sha256.hexdigest()[0:7]

    def _parse_identifier(self, identifier, payload):

        for rrset in payload["data"]:
            for record in rrset["records"]:
                if (
                    self._make_identifier(
                        rrset["type"], rrset["name"], record["content"]
                    )
                    == identifier
                ):
                    rtype = rrset["type"]
                    name = self._full_name(rrset["name"])
                    content = self._unclean_content(rrset["type"], record["content"])
                    return rtype, name, content

        raise Exception(f"Record with ID {identifier} not found ")

    def _clean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        elif rtype == "CNAME":
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            content = content.strip('"')
        elif rtype == "CNAME":
            content = self._full_name(content)
        return content



# pylint: disable=missing-docstring
import os

import pytest

import lexicon.client
from lexicon.config import ConfigResolver
from lexicon.tests.test_library import mock_fake_provider


@pytest.fixture(autouse=True)
def fake_provider():
    """Activate the fake_provider mock"""
    with mock_fake_provider():
        yield


def test_client_basic_init():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == options["domain"]
    assert client.config.resolve("lexicon:type") == options["type"]


def test_client_legacy_init(monkeypatch):
    monkeypatch.setenv("LEXICON_FAKEPROVIDER_AUTH_KEY", "MY_KEY")
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "example.com",
        "type": "TXT",
    }
    with pytest.deprecated_call():
        client = lexicon.client.Client(options)

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == options["domain"]
    assert client.config.resolve("lexicon:type") == options["type"]
    assert client.config.resolve("lexicon:fakeprovider:auth_key") == "MY_KEY"


def test_client_init_when_domain_includes_subdomain_should_strip():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "example.com"
    assert client.config.resolve("lexicon:type") == options["type"]


def test_client_init_with_delegated_domain_name():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.sub.example.com",
        "delegated": "sub",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "sub.example.com"
    assert client.config.resolve("lexicon:type") == options["type"]


def test_client_init_with_delegated_domain_fqdn():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.sub.example.com",
        "delegated": "sub.example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "sub.example.com"
    assert client.config.resolve("lexicon:type") == options["type"]


def test_client_init_with_same_delegated_domain_fqdn():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.example.com",
        "delegated": "example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options))

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "example.com"
    assert client.config.resolve("lexicon:type") == options["type"]


def test_client_init_when_missing_provider_should_fail():
    options = {"action": "list", "domain": "example.com", "type": "TXT"}
    with pytest.raises(AttributeError):
        lexicon.client.Client(ConfigResolver().with_dict(options))


def test_client_init_when_missing_action_should_fail():
    options = {"provider_name": "fakeprovider", "domain": "example.com", "type": "TXT"}
    with pytest.raises(AttributeError):
        lexicon.client.Client(ConfigResolver().with_dict(options))


def test_client_init_when_missing_domain_should_fail():
    options = {"provider_name": "fakeprovider", "action": "list", "type": "TXT"}
    with pytest.raises(AttributeError):
        lexicon.client.Client(ConfigResolver().with_dict(options))


def test_client_init_when_missing_type_should_fail():
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "example.com",
    }
    with pytest.raises(AttributeError):
        lexicon.client.Client(ConfigResolver().with_dict(options))


def test_client_parse_env_with_no_keys_should_do_nothing(monkeypatch):
    if os.environ.get("LEXICON_FAKEPROVIDER_TOKEN"):
        monkeypatch.delenv("LEXICON_FAKEPROVIDER_TOKEN")
    if os.environ.get("LEXICON_FAKEPROVIDER_USERNAME"):
        monkeypatch.delenv("LEXICON_FAKEPROVIDER_USERNAME")
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options).with_env())

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "example.com"
    assert client.config.resolve("lexicon:type") == options["type"]
    assert client.config.resolve("lexicon:fakeprovider:auth_token") is None
    assert client.config.resolve("lexicon:fakeprovider:auth_username") is None


def test_client_parse_env_with_auth_keys(monkeypatch):
    monkeypatch.setenv("LEXICON_FAKEPROVIDER_TOKEN", "test-token")
    monkeypatch.setenv("LEXICON_FAKEPROVIDER_USERNAME", "test-username@example.com")
    options = {
        "provider_name": "fakeprovider",
        "action": "list",
        "domain": "www.example.com",
        "type": "TXT",
    }
    client = lexicon.client.Client(ConfigResolver().with_dict(options).with_env())

    assert client.provider_name == options["provider_name"]
    assert client.action == options["action"]
    assert client.config.resolve("lexicon:domain") == "example.com"
    assert client.config.resolve("lexicon:type") == options["type"]
    assert client.config.resolve("lexicon:fakeprovider:auth_token") == "test-token"
    assert (
        client.config.resolve("lexicon:fakeprovider:auth_username")
        == "test-username@example.com"
    )


# TODO: add tests for Provider loading?



"""
This module takes care of finding information about the runtime of Lexicon:
* what are the providers installed, and available
* what is the version of Lexicon
"""
import pkgutil
from typing import Dict, List

import pkg_resources

from lexicon import providers


def find_providers() -> Dict[str, bool]:
    """Find all providers registered in Lexicon, and their availability"""
    providers_list = sorted(
        {
            modname
            for (_, modname, _) in pkgutil.iter_modules(providers.__path__)  # type: ignore
            if modname != "base"
        }
    )

    try:
        distribution = pkg_resources.get_distribution("dns-lexicon")
    except pkg_resources.DistributionNotFound:
        return {provider: True for provider in providers_list}
    else:
        return {
            provider: _resolve_requirements(provider, distribution)
            for provider in providers_list
        }


def lexicon_version() -> str:
    """Retrieve current Lexicon version"""
    try:
        return pkg_resources.get_distribution("dns-lexicon").version
    except pkg_resources.DistributionNotFound:
        return "unknown"


def _resolve_requirements(
    provider: str, distribution: pkg_resources.Distribution
) -> bool:
    try:
        requirements: List[pkg_resources.Requirement] = distribution.requires(
            extras=(provider,)
        )
    except pkg_resources.UnknownExtra:
        # No extra for this provider
        return True
    else:
        # Extra is defined
        try:
            for requirement in requirements:
                if hasattr(requirement, "name"):
                    pkg_resources.get_distribution(requirement.name)  # type: ignore
                else:
                    pkg_resources.get_distribution(requirement)
        except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict):
            # At least one extra requirement is not fulfilled
            return False

    return True



"""Module provider for Hover"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["hover.com"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Hover"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://www.hover.com/api"
        self.cookies = {}

    def _authenticate(self):
        # Getting required cookies "hover_session" and "hoverauth"
        response = requests.get("https://www.hover.com/signin")
        self.cookies["hover_session"] = response.cookies["hover_session"]

        payload = {
            "username": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        response = requests.post(
            "https://www.hover.com/api/login/", json=payload, cookies=self.cookies
        )
        response.raise_for_status()

        if "hoverauth" not in response.cookies:
            raise Exception("Unexpected auth response")
        self.cookies["hoverauth"] = response.cookies["hoverauth"]

        # Make sure domain exists
        # domain is stored in self.domain from BaseProvider

        domains = self._list_domains()
        for domain in domains:
            if domain["name"] == self.domain:
                self.domain_id = domain["id"]
                break
        else:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def _list_domains(self):
        response = self._get("/domains")

        domains = []
        for domain in response["domains"]:
            processed_domain = {
                "name": domain["domain_name"],
                "id": domain["id"],
                "active": (domain["status"] == "active"),
            }
            domains.append(processed_domain)

        LOGGER.debug("list_domains: %s", domains)
        return domains

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/domains/{self.domain_id}/dns")

        # payload['domains'] should be a list of len 1
        try:
            raw_records = payload["domains"][0]["entries"]
        except (KeyError, IndexError):
            raise Exception("Unexpected response")

        processed_records = []
        for record in raw_records:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            processed_records.append(processed_record)

        if rtype:
            processed_records = [
                record for record in processed_records if record["type"] == rtype
            ]
        if name:
            name = self._relative_name(name)
            processed_records = [
                record for record in processed_records if name in record["name"]
            ]
        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    def _create_record(self, rtype, name, content):
        name = self._relative_name(name)
        records = self._list_records(rtype, name, content)
        if records:
            LOGGER.debug("not creating duplicate record: %s", records[0])
            return True

        record = {"name": name, "type": rtype, "content": content}
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        LOGGER.debug("create_record: %s", record)
        payload = self._post(f"/domains/{self.domain_id}/dns", record)
        return payload["succeeded"]

    # Update a record. Hover cannot update name so we delete and recreate.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            records = [r for r in records if r["id"] == identifier]
        else:
            records = self._list_records(rtype, name, None)

        if not records:
            raise Exception("Record not found")
        if len(records) > 1:
            raise Exception("Record not unique")
        orig_record = records[0]
        orig_id = orig_record["id"]

        new_rtype = rtype if rtype else orig_record["type"]
        new_name = name if name else orig_record["name"]
        new_content = content if content else orig_record["content"]

        self._delete_record(orig_id)
        return self._create_record(new_rtype, new_name, new_content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_ids)

        for record_id in delete_record_ids:
            self._delete(f"/dns/{record_id}")
            LOGGER.debug("delete_record: %s", record_id)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            cookies=self.cookies,
            headers={"Content-Type": "application/json"},
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:  # response is not json
            raise Exception("Did not get JSON response.")



"""Integration tests for Rage4"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class Rage4ProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Rage4"""

    provider_name = "rage4"
    domain = "capsulecd.com"

    def _filter_headers(self):
        return ["Authorization"]

    @pytest.mark.skip(reason="update requires type to be specified for this provider")
    def test_provider_when_calling_update_record_with_full_name_should_modify_record(
        self,
    ):
        return

    @pytest.mark.skip(reason="update requires type to be specified for this provider")
    def test_provider_when_calling_update_record_should_modify_record(self):
        return

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(
        self,
    ):
        return



"""Module provider for DNSMadeEasy"""
import hmac
import json
import logging
from builtins import bytes
from email.utils import formatdate
from hashlib import sha1

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry  # type: ignore

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["dnsmadeeasy"]


class _RetryRateLimit(Retry):
    # Standard urllib3 Retry objects trigger retries only based on HTTP status code or HTTP method.
    # However we need to differentiate 400 errors with body `{"error": ["Rate limit exceeded"]}`
    # from the other 400 errors. The internal _RetryRateLimit class does that.
    def increment(
        self,
        method=None,
        url=None,
        response=None,
        error=None,
        _pool=None,
        _stacktrace=None,
    ):
        if response:
            body = json.loads(response.data)
            if "Rate limit exceeded" in body.get("error", []):
                return super(_RetryRateLimit, self).increment(
                    method, url, response, error, _pool, _stacktrace
                )

        raise RuntimeError(f"URL {url} returned a HTTP 400 status code.")


def provider_parser(subparser):
    """Configure provider parser for DNSMadeEasy"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument("--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for DNSMadeEasy"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint")
            or "https://api.dnsmadeeasy.com/V2.0"
        )

    def _authenticate(self):

        try:
            payload = self._get("/dns/managed/name", {"domainname": self.domain})
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                payload = {}
            else:
                raise

        if not payload or not payload["id"]:
            raise AuthenticationError("No domain found")

        self.domain_id = payload["id"]

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            "type": rtype,
            "name": self._relative_name(name),
            "value": content,
            "ttl": self._get_lexicon_option("ttl"),
        }
        payload = {}
        try:
            payload = self._post(f"/dns/managed/{self.domain_id}/records/", record)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code != 400:
                raise

                # http 400 is ok here, because the record probably already exists
        LOGGER.debug("create_record: %s", "name" in payload)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_query = {}
        if rtype:
            filter_query["type"] = rtype
        if name:
            filter_query["recordName"] = self._relative_name(name)
        payload = self._get(f"/dns/managed/{self.domain_id}/records", filter_query)

        records = []
        for record in payload["data"]:
            processed_record = {
                "type": record["type"],
                "name": f"{record['name']}.{self.domain}",
                "ttl": record["ttl"],
                "content": record["value"],
                "id": record["id"],
            }

            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {"id": identifier, "ttl": self._get_lexicon_option("ttl")}

        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["value"] = content
        if rtype:
            data["type"] = rtype

        self._put(f"/dns/managed/{self.domain_id}/records/{identifier}", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/dns/managed/{self.domain_id}/records/{record_id}")

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-dnsme-apiKey": self._get_provider_option("auth_username"),
        }
        default_auth = None

        # Date string in HTTP format e.g. Sat, 12 Feb 2011 20:59:04 GMT
        request_date = formatdate(usegmt=True)

        hashed = hmac.new(
            bytes(self._get_provider_option("auth_token"), "ascii"),
            bytes(request_date, "ascii"),
            sha1,
        )

        default_headers["x-dnsme-requestDate"] = request_date
        default_headers["x-dnsme-hmac"] = hashed.hexdigest()

        session = requests.Session()
        try:
            # DNSMadeEasy allows only 150 requests in a floating 5 min time window.
            # So we implement a retry strategy on requests returned as 400 with body
            # `{"error": ["Rate limit exceeded"]}`.
            # 10 retries with backoff = 0.6 gives following retry delays after first attempt:
            # 1.2s, 2.4s, 4.8s, 9.6s, 19.2s, 38.4s, 76.8s, 153.6s, 307.2s
            # So last attempt is done 5 min 7 seconds after first try, so the
            # size of the floating window.
            # Beyond it we can assume something else is wrong and so give up.
            session_retries = _RetryRateLimit(
                total=10, backoff_factor=0.6, status_forcelist=[400]
            )
            session_adapter = HTTPAdapter(max_retries=session_retries)
            session.mount("http://", session_adapter)
            session.mount("https://", session_adapter)
            response = session.request(
                action,
                self.api_endpoint + url,
                params=query_params,
                data=json.dumps(data),
                headers=default_headers,
                auth=default_auth,
            )
            # if the request fails for any reason, throw an error.
            response.raise_for_status()

            # PUT and DELETE actions dont return valid json.
            if action in ["DELETE", "PUT"]:
                return response.text
            return response.json()
        finally:
            session.close()



"""Test for one implementation of the interface"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class SafednsProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for SafeDNS provider"""

    provider_name = "safedns"
    domain = "lexicon.tests"

    def _filter_headers(self):
        return ["Authorization"]

    @pytest.mark.skip(reason="Record-level TTLs are not supported by this provider")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return

    @pytest.mark.skip(reason="CNAME requires FQDN for this provider")
    def test_provider_when_calling_create_record_for_CNAME_with_valid_name_and_content(
        self,
    ):
        return



from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
@pytest.mark.skip(reason="missing tests, will require to run Bind in docker")
class DDNSProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for DDNS provider"""

    provider_name = "DDNS"
    domain = "example.com"

    def _filter_post_data_parameters(self):
        return ["login_token"]

    def _filter_headers(self):
        return ["Authorization"]

    def _filter_query_parameters(self):
        return ["secret_key"]

    def _filter_response(self, response):
        """See `IntegrationTests._filter_response` for more information on how
        to filter the provider response."""
        return response



""""Test for rackspace implementation of the lexicon interface"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class RackspaceProviderTests(TestCase, IntegrationTestsV2):
    """Tests the rackspace provider"""

    provider_name = "rackspace"
    domain = "capsulecd.com"

    def _filter_post_data_parameters(self):
        return ["auth"]

    def _filter_headers(self):
        return ["X-Auth-Token"]

    # Rackspace does not provide a sandbox API; actual credentials are required
    # Replace the auth_account, auth_username and auth_api_key as well as the
    # domain above with an actual domain you have added to Rackspace to
    # regenerate the fixtures
    def _test_parameters_overrides(self):
        # Set this to 1 if you are making new recordings and set to 0 when
        # finished to make sure we don't actually sleep and waste time
        # when we are replaying. Rackspace API calls are async, so you place
        # the initial request and then make update calls to see if the action
        # is complete and has a response.
        return {"sleep_time": "0"}

    def _test_fallback_fn(self):
        return lambda x: "placeholder_" + x if x != "auth_token" else None



"""Test for route53 implementation of the interface."""
from contextlib import contextmanager
from unittest import TestCase

import pytest

from lexicon.tests.providers import integration_tests


class Route53ProviderTests(TestCase, integration_tests.IntegrationTestsV2):
    """Route53 Provider Tests."""

    provider_name = "route53"
    domain = "fullcr1stal.tk"

    def _filter_headers(self):
        """Sensitive headers to be filtered."""
        return ["Authorization"]

    def test_provider_authenticate_private_zone_only(self):
        with self._use_vcr("IntegrationTests/test_provider_authenticate.yaml"):
            provider = self._build_provider_with_overrides({"private_zone": "true"})
            with pytest.raises(Exception):
                provider.authenticate()

    def test_provider_authenticate_private_zone_false(self):
        with self._use_vcr("IntegrationTests/test_provider_authenticate.yaml"):
            provider = self._build_provider_with_overrides({"private_zone": "false"})
            provider.authenticate()
            assert provider.domain_id is not None

    def _build_provider_with_overrides(self, overrides):
        config = self._test_config()
        config.add_config_source(
            integration_tests.EngineOverrideConfigSource(overrides), 0
        )
        return self.provider_module.Provider(config)

    @contextmanager
    def _use_vcr(self, path):
        with integration_tests.PROVIDER_VCR.use_cassette(
            self._cassette_path(path),
            filter_headers=self._filter_headers(),
            before_record_response=self._filter_response,
            filter_query_parameters=self._filter_query_parameters(),
            filter_post_data_parameters=self._filter_post_data_parameters(),
        ):
            yield



"""Module provider for Mythic Beasts"""
import binascii
import json
import logging

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry  # type: ignore

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["mythic-beasts.com"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.description = """
        There are two ways to provide an authentication granting access to the Mythic Beasts API
        1 - With your API credentials (user/password),
            with --auth-username and --auth-password flags.
        2 - With an API token, using --auth-token flags.
        These credentials and tokens must be generated using the Mythic Beasts API v2.
    """
    subparser.add_argument(
        "--auth-username",
        help="specify API credentials username",
    )
    subparser.add_argument(
        "--auth-password",
        help="specify API credentials password",
    )
    subparser.add_argument(
        "--auth-token",
        help="specify API token for authentication",
    )


class Provider(BaseProvider):
    """Provider class for Mythic Beasts"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.mythic-beasts.com/dns/v2"
        self.auth_token = None

    def _authenticate(self):
        # may need to get auth token
        if self.auth_token is None and self._get_provider_option("auth_token") is None:
            auth_request = requests.request(
                "POST",
                "https://auth.mythic-beasts.com/login",
                data={"grant_type": "client_credentials"},
                auth=(
                    self._get_provider_option("auth_username"),
                    self._get_provider_option("auth_password"),
                ),
            )
            auth_request.raise_for_status()
            post_result = auth_request.json()

            if not post_result["access_token"]:
                raise AuthenticationError(
                    "Error, could not get access token "
                    f"for Mythic Beasts API for user: {self._get_provider_option('auth_username')}"
                )

            self.auth_token = post_result["access_token"]
        elif self.auth_token is None:
            self.auth_token = self._get_provider_option("auth_token")

        payload = self._get("/zones")

        if self.domain is None:
            if not payload["zones"]:
                raise AuthenticationError("No domain found")
            if len(payload["zones"]) > 1:
                raise AuthenticationError(
                    "Too many domains found. This should not happen"
                )
            else:
                self.domain = payload["zones"][0]
        else:
            if not payload["zones"]:
                raise AuthenticationError("No domain found")
            if self.domain not in payload["zones"]:
                raise AuthenticationError("Requested domain not found")

        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        LOGGER.debug("type %s", rtype)
        LOGGER.debug("name %s", name)
        LOGGER.debug("content %s", content)

        if rtype == "CNAME":
            content = self._fqdn_name(content)

        data = {
            "records": [
                {
                    "host": self._relative_name(name),
                    "type": rtype,
                    "data": content,
                }
            ]
        }
        if self._get_lexicon_option("ttl"):
            data["records"][0]["ttl"] = self._get_lexicon_option("ttl")

        payload = {"success": True}
        try:
            payload = self._post(f"/zones/{self.domain}/records", data)
        except requests.exceptions.HTTPError as err:
            if (
                err.response.status_code == 400
                and err.response.json()["errors"][0][0:16] == "Duplicate record"
            ):
                LOGGER.debug("create_record (ignored, duplicate)")
            else:
                raise

        if rtype == "A" or rtype == "TXT":
            # need to wait and poll here until verified that DNS change is live
            try:
                self._get(
                    f"/zones/{self.domain}/records/{self._relative_name(name)}/{rtype}?verify"
                )
            except requests.exceptions.HTTPError:
                LOGGER.debug("Timed out trying to verify changes were live")
                raise

        if "message" in payload:
            return payload["message"]
        elif "success" in payload:
            return payload["success"]

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_obj = {}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["host"] = self._relative_name(name)
        if content:
            filter_obj["data"] = content

        records = []
        payload = self._get(f"/zones/{self.domain}/records", filter_obj)

        LOGGER.debug("payload: %s", payload)

        for record in payload["records"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["host"]),
                "ttl": record["ttl"],
                "content": record["data"],
                # no id is available, so we need to make our own
                "id": _identifier(record),
            }
            if record["type"] == "MX" and record["mx_priority"]:
                processed_record["options"] = {
                    "mx": {"priority": record["mx_priority"]}
                }

            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        LOGGER.debug("Number of records retrieved: %d", len(records))
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self._list_records(rtype, self._full_name(name))
            if len(records) == 1:
                matching_record = records[0]
                filter_obj = {}
                filter_obj["type"] = matching_record["type"]
                filter_obj["host"] = self._relative_name(matching_record["name"])

            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )
        else:
            records = self._list_records()
            for record in records:
                if record["id"] == identifier:
                    matching_record = record
                    break
            else:
                raise Exception("Can't find record with that id!")

            filter_obj = {}
            filter_obj["type"] = matching_record["type"]
            filter_obj["host"] = self._relative_name(matching_record["name"])
            filter_obj["data"] = matching_record["content"]

        data = {"records": [{}]}
        if rtype:
            data["type"] = rtype
        if name:
            data["host"] = self._relative_name(name)
        if content:
            data["records"][0]["data"] = content
        if self._get_lexicon_option("ttl"):
            data["records"][0]["ttl"] = self._get_lexicon_option("ttl")

        LOGGER.debug(data)

        payload = self._put(
            f"/zones/{self.domain}/records/{self._relative_name(matching_record['name'])}/{matching_record['type']}",
            data,
            filter_obj,
        )

        LOGGER.debug("update_record: %s", payload["message"])
        return payload["message"]

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        filter_obj = {}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["host"] = self._relative_name(name)
        if content:
            filter_obj["data"] = content

        records = self._list_records(rtype, name, content)

        for record in records:
            LOGGER.debug("delete_records: %s", record)
            name = self._relative_name(record["name"])
            rtype = record["type"]
            if identifier is not None and identifier != record["id"]:
                continue

            self._delete(f"/zones/{self.domain}/records/{name}/{rtype}", filter_obj)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        # When editing DNS zone, results are not live immediately.
        # In this case, call to verify will return 409 HTTP error.
        # We use the Retry extension to retry the requests until
        # we get a processable response (402 HTTP status, or an HTTP error != 409)
        retries = Retry(
            total=10,
            backoff_factor=0.5,
            status_forcelist=[409],
            allowed_methods=frozenset(["GET", "PUT", "POST", "DELETE", "PATCH"]),
        )

        session = requests.Session()
        session.mount("https://", HTTPAdapter(max_retries=retries))

        headers = {"Content-Type": "application/json"}
        headers["Authorization"] = f"Bearer {self.auth_token}"
        response = session.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()


# Return hash id for record
def _identifier(record):
    digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
    digest.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    digest.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    digest.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

    return binascii.hexlify(digest.finalize()).decode("utf-8")[0:7]



"""Integration tests for auto"""
import socket
from unittest import TestCase, mock

import pytest

from lexicon.providers.auto import _get_ns_records_domains_for_domain
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# This fixture ensures to mock _get_ns_records_domains_for_domain, in order to not rely
# on the machine on which the test is done, as this function call nslookup.
# Then it will prevent errors where there is no network or tested domain do not exists anymore.
@pytest.fixture(autouse=True)
def _nslookup_mock(request):
    if request.node.name == "test_nslookup_resolution":
        # Do not mock for the test that specifically test nslookup resolution.
        yield
    else:
        with mock.patch(
            "lexicon.providers.auto._get_ns_records_for_domain",
            return_value=["ns.ovh.net"],
        ) as fixture:
            yield fixture


# Guys, are we online ?
def _there_is_no_network():
    try:
        socket.create_connection(("www.google.com", 80))
        return False
    except (OSError, IOError):
        pass
    return True


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class AutoProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for auto"""

    provider_name = "auto"
    domain = "pacalis.net"

    def _filter_headers(self):
        return ["X-Ovh-Application", "X-Ovh-Consumer", "X-Ovh-Signature"]

    def _test_parameters_overrides(self):
        return {"auth_entrypoint": "ovh-eu"}

    def _test_fallback_fn(self):
        return lambda x: "placeholder_" + x if x != "mapping_override" else None

    # Here we do not mock the function _get_ns_records_domains_for_domain
    # to effectively test the nslookup call and processing.\
    @pytest.mark.skipif(
        _there_is_no_network(), reason="No network, no nslookup call possible."
    )
    def test_nslookup_resolution(self):
        """Ensure that nameservers can be resolved through os nslookup call."""
        assert _get_ns_records_domains_for_domain("google.com")



"""Module provider for Dreamhost"""
import base64
import json
import logging
import time

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["dreamhost.com"]

_DATA_NON_EXIST_ERROR_LIST = [
    "no_record",
    "no_type",
    "no_value",
    "no_such_record",
    "no_such_type",
    "no_such_value",
    "no_such_zone",
]

_DATA_ALREADY_EXIST_ERROR_LIST = [
    "record_already_exists_not_editable",
    "record_already_exists_remove_first",
    "CNAME_already_on_record",
]


class NonExistError(Exception):
    """NonExistError"""


class AlreadyExistError(Exception):
    """AlreadyExistError"""


def provider_parser(subparser):
    """Module provider for Dreamhost"""
    subparser.add_argument("--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Dreamhost"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.dreamhost.com/"

    # Dreamhost provides no identifier for a record.
    # Furthermore, Dreamhost requires type, record, value to delete a record.
    # The record defined in lexicon is {type, name, content, id}
    # We use base64(json({'type', 'name', 'content'}))
    # as the identifier of Dreamhost record.
    @staticmethod
    def _identifier(dreamhost_record):
        id_struct = {
            "type": dreamhost_record["type"],
            "name": dreamhost_record["record"],
            "content": dreamhost_record["value"],
        }
        return base64.urlsafe_b64encode(json.dumps(id_struct).encode("utf-8")).decode(
            "utf-8"
        )

    # The information in identifier follows the record in lexicon.
    # Provider._record_to_dreamhost_record transfers to dreamhost-based record.
    @staticmethod
    def _id_to_dreamhost_record(identifier):
        record = json.loads(
            base64.urlsafe_b64decode(identifier.encode("utf-8")).decode("utf-8")
        )
        dreamhost_record = Provider._record_to_dreamhost_record(record)
        return dreamhost_record

    # The information in identifier follows the record in lexicon.
    # 'id' is added in the record.
    @staticmethod
    def _id_to_record(identifier):
        record = json.loads(
            base64.urlsafe_b64decode(identifier.encode("utf-8")).decode("utf-8")
        )
        record["id"] = identifier

        return record

    # Transferring lexicon-based record to Dreamhost-based record.
    @staticmethod
    def _record_to_dreamhost_record(record):
        dreamhost_record = {
            "type": record["type"],
            "record": record["name"],
            "value": record["content"],
        }
        return dreamhost_record

    def _authenticate(self):
        self.domain_id = None
        payload = self._get("dns-list_records")
        data = payload.get("data", None)
        if data is None:
            raise AuthenticationError("Domain not found")

        for record in data:
            if record.get("record", "") == self.domain and record.get("type", "") in [
                "A",
                "AAAA",
            ]:
                self.domain_id = self.domain
                break
        if self.domain_id is None:
            raise AuthenticationError("Domain not found")

    def _create_record(self, rtype, name, content):
        name = self._full_name(name)

        try:
            self._get(
                "dns-add_record",
                query_params={"record": name, "type": rtype, "value": content},
            )
        except AlreadyExistError:
            pass

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):

        payload = self._get("dns-list_records")

        resource_list = payload.get("data", None)
        if not isinstance(resource_list, list):
            raise Exception(f"unable to get records: {payload}")

        resource_list = [
            resource for resource in resource_list if resource["zone"] == self.domain
        ]
        if rtype:
            resource_list = [
                resource for resource in resource_list if resource["type"] == rtype
            ]
        if name:
            name = self._full_name(name)
            resource_list = [
                resource for resource in resource_list if resource["record"] == name
            ]
        if content:
            resource_list = [
                resource for resource in resource_list if resource["value"] == content
            ]

        processed_records = []
        for dreamhost_record in resource_list:
            processed_records.append(
                {
                    "id": Provider._identifier(dreamhost_record),
                    "type": dreamhost_record["type"],
                    "name": dreamhost_record["record"],
                    "content": dreamhost_record["value"],
                }
            )

        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            try:
                self._delete_record(identifier)
            except NonExistError:
                pass

        return self._create_record(rtype=rtype, name=name, content=content)

    # Delete existing records.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        to_deletes = []
        if identifier:
            record = Provider._id_to_record(identifier)
            to_deletes.append(record)
        else:
            records = self._list_records(rtype=rtype, name=name, content=content)
            to_deletes = records

        # for-loop to delete deletes.
        err = None
        for each in to_deletes:
            try:
                dreamhost_record = Provider._record_to_dreamhost_record(each)
                self._get("dns-remove_record", query_params=dreamhost_record)

            except Exception as exception:
                err = exception

            # Sleeping for 1-second to avoid trigerring ddos protecting in case of looped requests
            time.sleep(1)

        if err is not None:
            raise err

        return True

    # Helpers
    def _request(self, action="GET", url="", data=None, query_params=None):
        if data is None:
            data = {}

        if query_params is None:
            query_params = {}

        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        query_params["key"] = self._get_provider_option("auth_token")
        query_params["format"] = "json"
        if "cmd" not in query_params:
            query_params["cmd"] = url

        response = requests.request(
            action,
            self.api_endpoint,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        result = response.json()
        if result.get("result", "") != "success":
            err_msg = result.get("data", "")
            if err_msg in _DATA_NON_EXIST_ERROR_LIST:
                raise NonExistError(f"Dreamhost non-exist error: {result}")
            if err_msg in _DATA_ALREADY_EXIST_ERROR_LIST:
                raise AlreadyExistError(f"Dreamhost already-exist error: {result}")
            raise Exception(f"Dreamhost api error: {result}")
        return result



"""Module provider for Netcup"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["netcup.de"]


def provider_parser(subparser):
    """Configure provider parser for Netcup"""
    subparser.add_argument(
        "--auth-customer-id", help="specify customer number for authentication"
    )
    subparser.add_argument("--auth-api-key", help="specify API key for authentication")
    subparser.add_argument(
        "--auth-api-password", help="specify API password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Netcup"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = 0
        self.zone_ttl = None
        self.api_session_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint")
            or "https://ccp.netcup.net/run/webservice/servers/endpoint.php?JSON"
        )

    # lexicon.provider.Provider overrides:

    def _authenticate(self):
        """Authenticate with netcup server. Must be called first."""
        login_info = self._apicall("login")
        self.api_session_id = login_info["apisessionid"]
        if not self.api_session_id:
            raise AuthenticationError("Login failed")
        # query ttl and verify access to self.domain:
        zone_info = self._apicall("infoDnsZone", domainname=self.domain)
        self.zone_ttl = zone_info["ttl"]

    def _create_record(self, rtype, name, content):
        """Create record. If it already exists, do nothing."""
        if not self._list_records(rtype, name, content):
            self._update_records(
                [{}],
                {
                    "type": rtype,
                    "hostname": self._relative_name(name),
                    "destination": content,
                    "priority": self._get_lexicon_option("priority"),
                },
            )
        LOGGER.debug("create_record: %s", True)
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        """List all records. Return an empty list if no records found.
        ``rtype``, ``name`` and ``content`` are used to filter records."""
        records = [
            {
                "id": record["id"],
                "type": record["type"],
                "name": self._full_name(record["hostname"]),
                "content": record["destination"],
                "priority": record["priority"],
                "ttl": self.zone_ttl,
            }
            for record in self._raw_records(None, rtype, name, content)
        ]
        LOGGER.debug("list_records: %s", records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """Create or update a record."""
        records = self._raw_records(identifier, rtype, name, content)
        self._update_records(
            records,
            {
                "type": rtype,
                "hostname": self._relative_name(name),
                "destination": content,
                "priority": self._get_lexicon_option("priority"),
            },
        )
        LOGGER.debug("update_record: %s", True)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Delete an existing record. If record does not exist, do nothing."""
        records = self._raw_records(identifier, rtype, name, content)
        LOGGER.debug("delete_records: %s", [rec["id"] for rec in records])
        self._update_records(
            records,
            {
                "deleterecord": True,
                "type": rtype,
                "hostname": name,
                "destination": content,
            },
        )
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _raw_records(self, identifier=None, rtype=None, name=None, content=None):
        """Return list of record dicts in the netcup API convention."""
        record_fields = {
            "id": identifier,
            "type": rtype,
            "hostname": name and self._relative_name(name),
            "destination": content,
        }
        # type/hostname/destination of the dnsrecord type are mandatory (even
        # when deleting), and must be queried if not all were specified:
        if all(record_fields.values()):
            return [record_fields]
        data = self._apicall("infoDnsRecords", domainname=self.domain)
        records = data.get("dnsrecords", [])
        return [
            record
            for record in records
            if all(record[k] == v for k, v in record_fields.items() if v)
        ]

    def _update_records(self, records, data):
        """Insert or update a list of DNS records, specified in the netcup API
        convention.

        The fields ``hostname``, ``type``, and ``destination`` are mandatory
        and must be provided either in the record dict or through ``data``!
        """
        data = {k: v for k, v in data.items() if v}
        records = [dict(record, **data) for record in records]
        return self._apicall(
            "updateDnsRecords",
            domainname=self.domain,
            dnsrecordset={"dnsrecords": records},
        ).get("dnsrecords", [])

    def _apicall(self, method, **params):
        """Call an API method and return response data. For more info, see:
        https://ccp.netcup.net/run/webservice/servers/endpoint"""
        LOGGER.debug("%s(%r)", method, params)
        auth = {
            "customernumber": self._get_provider_option("auth_customer_id"),
            "apikey": self._get_provider_option("auth_api_key"),
        }
        if method == "login":
            auth["apipassword"] = self._get_provider_option("auth_api_password")
        else:
            auth["apisessionid"] = self.api_session_id
        if not all(auth.values()):
            raise Exception("No valid authentication mechanism found")
        data = self._request(
            "POST", url="", data={"action": method, "param": dict(params, **auth)}
        )
        if data["status"] != "success":
            raise Exception(f"{data['longmessage']} ({data['statuscode']})")
        return data.get("responsedata", {})

    def _request(self, action="GET", url="/", data=None, query_params=None):
        """Perform network request to configured JSON endpoint."""
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        response = requests.request(
            action,
            self.api_endpoint + url,
            data=json.dumps(data),
            params=query_params,
            headers=default_headers,
        )
        response.raise_for_status()
        return response.json()



"""Module provider for Transip"""
import logging
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

# Support various versions of Transip Python API
try:
    from transip.service.objects import DnsEntry  # type: ignore
except ImportError:
    try:
        from transip.service.dns import DnsEntry  # type: ignore
    except ImportError:
        pass

try:
    from transip.service.domain import DomainService  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS: List[str] = []


def provider_parser(subparser):
    """Configure provider parser for Transip"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-api-key", help="specify API private key for authentication"
    )


class Provider(BaseProvider):
    """
    Provider class for Transip

    provider_options can be overwritten by a Provider to setup custom defaults.
    They will be overwritten by any options set via the CLI or Env.
    order is:

    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.provider_name = "transip"
        self.domain_id = None

        username = self._get_provider_option("auth_username")
        key_file = self._get_provider_option("auth_api_key")

        if not username or not key_file:
            raise Exception("No username and/or keyfile was specified")

        self.client = DomainService(login=username, private_key_file=key_file)

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def _authenticate(self):
        # This request will fail when the domain does not exist,
        # allowing us to check for existence
        try:
            self.client.get_info(self.domain)
        except BaseException:
            raise AuthenticationError(
                f"Could not retrieve information about {self.domain}, is this domain yours?"
            )
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        records = self.client.get_info(self.domain).dnsEntries

        if self._filter_records(records, rtype, name, content):
            # Nothing to do, record already exists
            LOGGER.debug("create_record: already exists")
            return True

        records.append(
            DnsEntry(
                **{
                    "name": self._relative_name(name),
                    "record_type": rtype,
                    "content": self._bind_format_target(rtype, content),
                    "expire": self._get_lexicon_option("ttl"),
                }
            )
        )

        self.client.set_dns_entries(self.domain, records)
        status = (
            len(self._list_records_internal(rtype, name, content, show_output=False))
            >= 1
        )
        LOGGER.debug("create_record: %s", status)
        return status

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, show_output=True
    ):
        all_records = self._convert_records(
            self.client.get_info(self.domain).dnsEntries
        )
        records = self._filter_records(
            records=all_records, rtype=rtype, name=name, content=content
        )

        if show_output:
            LOGGER.debug("list_records: %s", records)
        return records

    # Update a record. Identifier must be specified.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not (rtype or name or content):
            raise Exception("At least one of rtype, name or content must be specified.")

        all_records = self._list_records_internal(show_output=False)
        filtered_records = self._filter_records(all_records, rtype, name)

        for record in filtered_records:
            all_records.remove(record)
        all_records.append(
            {
                "name": name,
                "type": rtype,
                "content": self._bind_format_target(rtype, content),
                "ttl": self._get_lexicon_option("ttl"),
            }
        )

        self.client.set_dns_entries(
            self.domain, self._convert_records_back(all_records)
        )
        status = (
            len(self._list_records_internal(rtype, name, content, show_output=False))
            >= 1
        )
        LOGGER.debug("update_record: %s", status)
        return status

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not (rtype or name or content):
            raise Exception("At least one of rtype, name or content must be specified.")

        all_records = self._list_records_internal(show_output=False)
        filtered_records = self._filter_records(all_records, rtype, name, content)

        for record in filtered_records:
            all_records.remove(record)

        self.client.set_dns_entries(
            self.domain, self._convert_records_back(all_records)
        )
        status = (
            len(self._list_records_internal(rtype, name, content, show_output=False))
            == 0
        )
        LOGGER.debug("delete_record: %s", status)
        return status

    def _full_name(self, record_name):
        if record_name == "@":
            record_name = self.domain
        return super(Provider, self)._full_name(record_name)

    def _relative_name(self, record_name):
        name = super(Provider, self)._relative_name(record_name)
        if not name:
            name = "@"
        return name

    def _bind_format_target(self, rtype, target):
        if rtype == "CNAME" and not target.endswith("."):
            target += "."
        return target

    # Convert the objects from transip to dicts, for easier processing
    def _convert_records(self, records):
        _records = []
        for record in records:
            _records.append(
                {
                    "id": f"{self._full_name(record.name)}-{record.type}",
                    "name": self._full_name(record.name),
                    "type": record.type,
                    "content": record.content,
                    "ttl": record.expire,
                }
            )
        return _records

    def _to_dns_entry(self, _entry):
        return DnsEntry(
            self._relative_name(_entry["name"]),
            _entry["ttl"],
            _entry["type"],
            _entry["content"],
        )

    def _convert_records_back(self, _records):
        return [self._to_dns_entry(record) for record in _records]

    # Filter a list of records based on criteria
    def _filter_records(self, records, rtype=None, name=None, content=None):
        _records = []
        for record in records:
            if (
                (not rtype or record["type"] == rtype)
                and (
                    not name or self._full_name(record["name"]) == self._full_name(name)
                )
                and (not content or record["content"] == content)
            ):
                _records.append(record)
        return _records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in Transip.
        pass



"""Integration tests for PowerDNS"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class PowerdnsProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for PowerDNS"""

    provider_name = "powerdns"
    domain = "sometestdomain.com"

    def _filter_headers(self):
        return ["X-API-Key"]

    def _test_parameters_overrides(self):
        return {"pdns_server": "http://127.0.0.1:8081", "pdns_server_id": "localhost"}

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(
        self,
    ):
        return



"""Module provider for Namesilo"""
import logging
from xml.etree import ElementTree

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["namesilo.com"]


def provider_parser(subparser):
    """Configure provider parser for Namesilo"""
    subparser.add_argument("--auth-token", help="specify key for authentication")


class Provider(BaseProvider):
    """Provider class for Namesilo"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint") or "https://www.namesilo.com/api"
        )

    def _authenticate(self):
        try:
            self._get("/getDomainInfo", {"domain": self.domain})
        except Exception as e:
            raise AuthenticationError(str(e))
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            "domain": self.domain_id,
            "rrhost": self._relative_name(name),
            "rrtype": rtype,
            "rrvalue": content,
        }
        if self._get_lexicon_option("ttl"):
            record["rrttl"] = self._get_lexicon_option("ttl")
        try:
            self._get("/dnsAddRecord", record)
        except ValueError as err:
            # noop if attempting to create record that already exists.
            LOGGER.debug("Ignoring error: %s", err)

        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        query = {"domain": self.domain_id}

        payload = self._get("/dnsListRecords", query)
        records = []
        for record in payload.find("reply").findall("resource_record"):
            processed_record = {
                "type": record.find("type").text,
                "name": record.find("host").text,
                "ttl": record.find("ttl").text,
                "content": record.find("value").text,
                "id": record.find("record_id").text,
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {"domain": self.domain_id, "rrid": identifier}
        # if rtype:
        #     data['type'] = rtype
        if name:
            data["rrhost"] = self._relative_name(name)
        if content:
            data["rrvalue"] = content
        if self._get_lexicon_option("ttl"):
            data["rrttl"] = self._get_lexicon_option("ttl")

        self._get("/dnsUpdateRecord", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        data = {"domain": self.domain_id}

        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            data["rrid"] = record_id
            self._get("/dnsDeleteRecord", data)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        query_params["version"] = 1
        query_params["type"] = "xml"
        query_params["key"] = self._get_provider_option("auth_token")
        response = requests.request(
            action, self.api_endpoint + url, params=query_params
        )
        # data=json.dumps(data))
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        tree = ElementTree.ElementTree(ElementTree.fromstring(response.content))
        root = tree.getroot()
        if root.find("reply").find("code").text == "280":
            raise ValueError(
                "An error occurred: {0}, {1}".format(
                    root.find("reply").find("detail").text,
                    root.find("reply").find("code").text,
                )
            )
        if root.find("reply").find("code").text != "300":
            raise Exception(
                "An error occurred: {0}, {1}".format(
                    root.find("reply").find("detail").text,
                    root.find("reply").find("code").text,
                )
            )

        return root



"""Integration tests for DNSSimple"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class DnsimpleProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for DNSSimple"""

    provider_name = "dnsimple"
    domain = "lexicontest.us"

    def _test_parameters_overrides(self):
        return {
            "api_endpoint": "https://api.sandbox.dnsimple.com/v2",
            "region": "global",
        }

    def _filter_headers(self):
        return ["Authorization", "set-cookie", "X-Dnsimple-OTP"]



"""Integration tests for Online.net"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class OnlineProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Online.net"""

    provider_name = "online"
    domain = "capsulecd.com"

    def _filter_headers(self):
        return ["Authorization", "x-recruitment"]

    def _test_fallback_fn(self):
        return lambda x: "placeholder_" + x if x != "priority" else ""

    @pytest.mark.skip(reason="manipulating records by id is not supported")
    def test_provider_when_calling_delete_record_by_identifier_should_remove_record(
        self,
    ):
        return



"""Integration tests for CloudNS"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class CloudnsProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for CloudNS"""

    provider_name = "cloudns"
    domain = "api-example.com"

    def _filter_query_parameters(self):
        return ["auth-id", "sub-auth-id", "sub-auth-user", "auth-password"]

    def _filter_post_data_parameters(self):
        return ["auth-id", "sub-auth-id", "sub-auth-user", "auth-password"]



"""Integration tests for DirectAdmin"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse the
# tests which *each and every* implementation of the interface must pass, by
# inheritance from integration_tests.IntegrationTests
class DirectAdminProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for DirectAdmin provider"""

    provider_name = "directadmin"
    domain = "example.com"
    endpoint = "http://localhost"

    def _filter_post_data_parameters(self):
        return []

    def _filter_headers(self):
        return ["Authorization"]

    def _filter_query_parameters(self):
        return []

    def _filter_response(self, response):
        """See `IntegrationTests._filter_response` for more information on how
        to filter the provider response."""
        return response

    def _test_parameters_overrides(self):
        return {"endpoint": self.endpoint}



# coding: utf-8
# Copyright (c) 2016, Jason Kulatunga
# Copyright (c) 2021, Oracle and/or its affiliates.
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class OciProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for OCI provider"""

    provider_name = "oci"
    domain = "lexicon-test.com"

    # Provide a fake credentials but the same region as the recorded tests
    def _test_parameters_overrides(self):
        return {
            "auth_user": "ocid1.user.oc1..aaaaaaa",
            "auth_tenancy": "ocid1.tenancy.oc1..aaaaaaaa",
            "auth_fingerprint": "aa:11:bb:22:cc:33:dd:44:ee:55:ff:66:a1:77:a2:88",
            "auth_region": "us-ashburn-1",
            "auth_key_content": """-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEA17lz4CMII+rpvu9Pi/AkOosjQR+kvDxdpbbxYY2JMkRnTuuN
nb7kefWQ9fEqsmTJcaUl0rrDORrYU+W+WYmuPYh839BeKPFbg/WbfPFMhLwF2X6G
w2Ozj7a9JceEYsuOptxtYawYabsN3wbPLAuTvY4B08kGLgVIzs7A3L/6JPbHJhjV
YuiRwWe2zzD64ERe3UuGEOXcTyU/guS3bX91W46E4vw6CFL2GiNgI5Q46xF3D31/
P/GKdFewm8rr7kAEouGW4mhOw0KI/mqgNrctQt+YQAzw/QKI4+umMu01ZtuYXRbF
40JsTuQOxnmClf7P59YrsI3ltBA94XthyVxh0QIDAQABAoIBADqnXt0zSTRS2+kh
MjSvP3p3eEdtriHMG/5BppHKpOH4/UnU+/VHAOI0JYzpXJ2Sj78JkyYfx5LQPL9a
+Q1pROnQIXvNMLzbGvHfJr6q8Q4p/UEsiMG5awoJOpZ6EAG4rPmrd0YWP7EHvfbE
6DFmmG3ynYaS4s5Ce5BXYNLkk8PWoWpEBGDOrdOU6bvMFqEgzqjnlbVNwGDP09HQ
fGtnycAqOsyC2CnrpNoOq6TWr/Y7s0EViajYxxTgoSgXnscntz8M2tqCW3Hd1T94
w2DOe1W5EGIkVrNpTePwFCkqPI47kTUT1eHDowf5NmRADJVDHr5H6yNCk7d/u2mr
6JsggE0CgYEA8YUBc/fagXLVAsEi3sOcyJrq04C+oy3Zz7zHz7ZH72cfJvaPlBPI
FPLVhEptn7bPMR7JuzUL6BckEnoPdXG52ncnZZbKMKXfBS6IkC/SSYjX6TWBXH9I
523QUi45N68UGb+opX/0iTvSrimK4wDWdz+Umx3rbt6uqmwuCHj9dn8CgYEA5KiG
7bnMRQAMadOOeFQePpViisEuHAAtL4f8qbXrwEjEjCjFBb3Hifdl6bRsvU8LEcN2
6YUM+KQXTOu+10N+4ANptjXzHJMIliO6FQ+7S8Rs/wDSrtuKHaEQR19HgyJzPfSi
ptZNWBFphrFE3nYO2GoYqCGwFYqNkW084XPBH68CgYEAyZcJBXEF0zK0FV576pA/
1zlndC5r8Owed8TMytUM6gia+fynDyQLx2CBU7CEG+GMwyU9oKLAU3KtSzbSnGbW
iEEYgzT/gueQZVTX6/Hehj5QaXmdhkU/5tvEHDQ00gOytWNCMxHAXKOwUGqgYKWc
XWCWe3rXvmzkQZ+WNMA4X6UCgYEAgJcmClsKvWMhmAIZhSIJQDjSiiXJwIV449oe
BXMBeclyf0AOTQRFSxmOfrewz2W8W+kI3pqsiMf/MosBcB3NJD3HHWmJpvApTAYb
h+yo8BsvENltolhke/UwKnMyzFR7asRBFIJATN698bmPeWv7PUmtRCBt3i9lHfvI
2SE34pECgYABPlBhx/irQMLXOacZxpXdVUj6/ivy4asiHe4OZN4uSXS1WqmrRHBe
Spzc4VkUFCDazQqxbG2/4ElGUzw+z0C0KMsSfSXTIk/YpNkyrCTwsllyvZrH7gVP
qpq9aJr7G9qW+9Yovwm2qLR1joa4XoOqn13bkEU7lc96fyRBr0ucDg==
-----END RSA PRIVATE KEY-----
""",
        }

    def _filter_headers(self):
        return ["authorization", "x-content-sha256"]

    def _filter_response(self, response):

        response["body"]["string"] = re.sub(
            br'"compartmentId":"[\w.-]+"',
            b'"compartmentId":"OCI-COMPARTMENT-ID"',
            response["body"]["string"],
        )
        response["body"]["string"] = re.sub(
            br'"id":"[\w.-]+"',
            b'"id":"DNS-ZONE-ID"',
            response["body"]["string"],
        )
        return response



"""Integration tests for Gehirn"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class GehirnProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for Gehirn"""

    provider_name = "gehirn"
    domain = "example.com"

    def _filter_headers(self):
        return ["Authorization"]



"""Provide support to Lexicon for Gandi LiveDNS and Gandi XMLRPC changes.

Lexicon provides a common interface for querying and managing DNS services
through those services' APIs. This module implements the Lexicon interface
against the Gandi API.

Gandi introduced the LiveDNS API (http://doc.livedns.gandi.net/) in 2017.
It is the successor of the traditional XMLRPC API, which suffered from
long delays between API-based changes and their activation.
The LiveDNS API has one significant peculiarity: DNS records with the
same name and type are managed as one unit. Thus records cannot be
addressed distinctly, which complicates removal of records significantly.

The Gandi XMLRPC API is different from typical DNS APIs in that Gandi
zone changes are atomic. You cannot edit the currently active
configuration. Any changes require editing either a new or inactive
configuration. Once the changes are committed, then the domain is switched
to using the new zone configuration. This module makes no attempt to
cleanup previous zone configurations.

Note that Gandi domains can share zone configurations. In other words,
I can have domain-a.com and domain-b.com which share the same zone
configuration file. If I make changes to domain-a.com, those changes
will only apply to domain-a.com, as domain-b.com will continue using
the previous version of the zone configuration. This module makes no
attempt to detect and account for that.
"""
import json
import logging
from builtins import object

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

try:
    import xmlrpclib  # type: ignore
except ImportError:
    import xmlrpc.client as xmlrpclib  # type: ignore

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["gandi.net"]


def provider_parser(subparser):
    """Specify arguments for Gandi Lexicon Provider."""
    subparser.add_argument("--auth-token", help="specify Gandi API key")
    subparser.add_argument(
        "--api-protocol",
        help="(optional) specify Gandi API protocol to use: rpc (default) or rest",
    )


class Provider(BaseProvider):
    """Provide Gandi LiveDNS API implementation of Lexicon Provider interface.

    Note that this implementation will delegates its call to GandiRPCSubProvider
    if RPC protocol is used.
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.default_ttl = 3600
        self.protocol = self._get_provider_option("api_protocol") or "rpc"

        if self.protocol != "rpc" and self.protocol != "rest":
            raise ValueError(
                "Invalid API protocol specified, should be 'rpc' or 'rest'"
            )

        if self.protocol == "rpc":
            self.rpc_helper = GandiRPCSubProvider(
                self._get_provider_option("auth_token"),
                "https://rpc.gandi.net/xmlrpc/",
                self.domain.lower(),
                self._relative_name,
                self._full_name,
            )
        else:
            self.api_endpoint = "https://api.gandi.net/v5/livedns"

    def _authenticate(self):
        if self.protocol == "rpc":
            domain_id = self.rpc_helper.authenticate()
            self.domain_id = domain_id
        else:
            self._get(f"/domains/{self.domain}")
            self.domain_id = self.domain.lower()

    def _create_record(self, rtype, name, content):
        if self.protocol == "rpc":
            return self.rpc_helper.create_record(
                rtype,
                self._relative_name(name),
                content,
                self._get_lexicon_option("ttl") or self.default_ttl,
            )

        current_values = [
            record["content"] for record in self._list_records(rtype=rtype, name=name)
        ]
        if current_values != [content]:
            # a change is necessary
            url = (
                f"/domains/{self.domain_id}/records/{self._relative_name(name)}/{rtype}"
            )
            if current_values:
                record = {"rrset_values": current_values + [content]}
                if self._get_lexicon_option("ttl"):
                    record["rrset_ttl"] = (
                        self._get_lexicon_option("ttl")
                        if self._get_lexicon_option("ttl") >= 300
                        else 300
                    )
                self._put(url, record)
            else:
                record = {"rrset_values": [content]}
                # add the ttl, if this is a new record
                if self._get_lexicon_option("ttl"):
                    record["rrset_ttl"] = (
                        self._get_lexicon_option("ttl")
                        if self._get_lexicon_option("ttl") >= 300
                        else 300
                    )
                self._post(url, record)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        """List all record for the domain in the active Gandi zone."""
        if self.protocol == "rpc":
            return self.rpc_helper.list_records(rtype, name, content)

        try:
            if name is not None:
                if rtype is not None:
                    query_results = [
                        self._get(
                            f"/domains/{self.domain_id}/records/{self._relative_name(name)}/{rtype}"
                        )
                    ]
                else:
                    query_results = self._get(
                        f"/domains/{self.domain_id}/records/{self._relative_name(name)}"
                    )
            else:
                query_results = self._get(f"/domains/{self.domain_id}/records")
                if rtype is not None:
                    query_results = [
                        item for item in query_results if item["rrset_type"] == rtype
                    ]
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                query_results = []
            else:
                raise

        # convert records with multiple values into single-value records
        records = []
        for query_result in query_results:
            for value in query_result["rrset_values"]:
                record = {
                    "type": query_result["rrset_type"],
                    "name": self._full_name(query_result["rrset_name"]),
                    "ttl": query_result["rrset_ttl"],
                    "content": value,
                    "id": query_result["rrset_name"],
                }
                # cleanup potential quoting if suitable
                self._clean_TXT_record(record)
                records.append(record)
        # filter for content, if requested
        if content is not None:
            records = [record for record in records if record["content"] == content]
        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record. Identifier must be specified.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """Updates the specified record in a new Gandi zone

        'content' should be a string or a list of strings
        """
        if self.protocol == "rpc":
            return self.rpc_helper.update_record(identifier, rtype, name, content)

        data = {}
        if rtype:
            data["rrset_type"] = rtype
        if name:
            data["rrset_name"] = self._relative_name(name)
        if content:
            if isinstance(content, (list, tuple, set)):
                data["rrset_values"] = list(content)
            else:
                data["rrset_values"] = [content]
        if rtype is not None:
            # replace the records of a specific rtype
            effect_id = identifier or self._relative_name(name)
            url = f"/domains/{self.domain_id}/records/{effect_id}/{rtype}"
            self._put(url, data)
        else:
            # replace all records with a matching name
            effect_id = identifier or self._relative_name(name)
            url = f"/domains/{self.domain_id}/records/{effect_id}"
            self._put(url, {"items": [data]})
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete existings records.
    # If records do not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if self.protocol == "rpc":
            return self.rpc_helper.delete_record(identifier, rtype, name, content)

        if not identifier:
            remove_count = 0
            # get all matching (by rtype and name) records - ignore 'content' for now
            records = self._list_records(rtype=rtype, name=name)
            for current_type in set(record["type"] for record in records):
                matching_records = [
                    record for record in records if record["type"] == current_type
                ]
                # collect all non-matching values
                if content is None:
                    remaining_values = []
                else:
                    remaining_values = [
                        record["content"]
                        for record in matching_records
                        if record["content"] != content
                    ]
                url = f"/domains/{self.domain_id}/records/{self._relative_name(name)}/{current_type}"
                if len(matching_records) == len(remaining_values):
                    # no matching item should be removed for this rtype
                    pass
                elif remaining_values:
                    # reduce the list of values
                    self._put(url, {"rrset_values": remaining_values})
                    remove_count += 1
                else:
                    # remove the complete record (possibly with multiple values)
                    self._delete(url)
                    remove_count += 1

            if remove_count == 0:
                raise Exception("Record identifier could not be found.")
        else:
            self._delete(f"/domains/{self.domain_id}/records/{identifier}")

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Apikey " + self._get_provider_option("auth_token"),
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        return response.json()


class GandiRPCSubProvider(object):
    """Provide Gandi RPCXML API implementation of Lexicon Provider interface.
    This implementation is called through the main LiveDNS implementation
    is RPC protocol is used.
    """

    def __init__(self, api_key, api_endpoint, domain, relative_name_fn, full_name_fn):
        """Initialize Gandi RCPXML API provider."""
        super(GandiRPCSubProvider, self).__init__()

        self._api_endpoint = api_endpoint
        self._api_key = api_key
        self._domain = domain
        self._relative_name = relative_name_fn
        self._full_name = full_name_fn
        self._api = xmlrpclib.ServerProxy(self._api_endpoint, allow_none=True)
        self._zone_id = None

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls. Should throw an error if
    # authentication fails for any reason, or if the domain does not exist.
    def authenticate(self):
        """Determine the current domain and zone IDs for the domain."""
        try:
            payload = self._api.domain.info(self._api_key, self._domain)
            self._zone_id = payload["zone_id"]
            return payload["id"]
        except xmlrpclib.Fault as err:
            raise AuthenticationError(f"Failed to authenticate: '{err}'")

    # Create record. If record already exists with the same content, do nothing.
    def create_record(self, rtype, name, content, ttl):
        """Creates a record for the domain in a new Gandi zone."""
        version = None
        ret = False

        # This isn't quite "do nothing" if the record already exists.
        # In this case, no new record will be created, but a new zone version
        # will be created and set.
        try:
            version = self._api.domain.zone.version.new(self._api_key, self._zone_id)
            self._api.domain.zone.record.add(
                self._api_key,
                self._zone_id,
                version,
                {"type": rtype.upper(), "name": name, "value": content, "ttl": ttl},
            )
            self._api.domain.zone.version.set(self._api_key, self._zone_id, version)
            ret = True

        finally:
            if not ret and version is not None:
                self._api.domain.zone.version.delete(
                    self._api_key, self._zone_id, version
                )

        LOGGER.debug("create_record: %s", ret)
        return ret

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        """List all record for the domain in the active Gandi zone."""
        opts = {}
        if rtype is not None:
            opts["type"] = rtype.upper()
        if name is not None:
            opts["name"] = self._relative_name(name)
        if content is not None:
            opts["value"] = (
                self._txt_encode(content) if opts.get("type", "") == "TXT" else content
            )

        records = []
        payload = self._api.domain.zone.record.list(
            self._api_key, self._zone_id, 0, opts
        )
        for record in payload:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["value"],
                "id": record["id"],
            }

            # Gandi will add quotes to all TXT record strings
            if processed_record["type"] == "TXT":
                processed_record["content"] = self._txt_decode(
                    processed_record["content"]
                )

            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record. Identifier or type+name+content
    def update_record(self, identifier, rtype=None, name=None, content=None):
        """Updates the specified record in a new Gandi zone."""
        if not identifier:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) > 1:
                raise Exception("Several record identifiers match the request")
            else:
                raise Exception("Record identifier could not be found")

        identifier = str(identifier)
        version = None

        # Gandi doesn't allow you to edit records on the active zone file.
        # Gandi also doesn't persist zone record identifiers when creating
        # a new zone file. To update by identifier, we lookup the record
        # by identifier, then use the record fields to find the record in
        # the newly created zone.
        records = self._api.domain.zone.record.list(
            self._api_key, self._zone_id, 0, {"id": identifier}
        )

        if len(records) == 1:
            rec = records[0]
            del rec["id"]

            try:
                version = self._api.domain.zone.version.new(
                    self._api_key, self._zone_id
                )
                records = self._api.domain.zone.record.list(
                    self._api_key, self._zone_id, version, rec
                )
                if len(records) != 1:
                    raise self.GandiInternalError("expected one record")

                if rtype is not None:
                    rec["type"] = rtype.upper()
                if name is not None:
                    rec["name"] = self._relative_name(name)
                if content is not None:
                    rec["value"] = (
                        self._txt_encode(content) if rec["type"] == "TXT" else content
                    )

                records = self._api.domain.zone.record.update(
                    self._api_key, self._zone_id, version, {"id": records[0]["id"]}, rec
                )
                if len(records) != 1:
                    raise self.GandiInternalError("Expected one updated record")

                self._api.domain.zone.version.set(self._api_key, self._zone_id, version)
                ret = True

            except self.GandiInternalError:
                pass

            finally:
                if not ret and version is not None:
                    self._api.domain.zone.version.delete(
                        self._api_key, self._zone_id, version
                    )

        LOGGER.debug("update_record: %s", ret)
        return ret

    # Delete existing records.
    # If records do not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Removes the specified records in a new Gandi zone."""
        version = None
        ret = False

        opts = {}
        if identifier is not None:
            opts["id"] = identifier
        else:
            if not rtype and not name and not content:
                raise ValueError(
                    "Error, at least one parameter from type, name or content must be set"
                )
            if rtype:
                opts["type"] = rtype.upper()
            if name:
                opts["name"] = self._relative_name(name)
            if content:
                opts["value"] = (
                    self._txt_encode(content) if opts["type"] == "TXT" else content
                )

        records = self._api.domain.zone.record.list(
            self._api_key, self._zone_id, 0, opts
        )

        if records:
            try:
                version = self._api.domain.zone.version.new(
                    self._api_key, self._zone_id
                )
                for record in records:
                    del record["id"]
                    self._api.domain.zone.record.delete(
                        self._api_key, self._zone_id, version, record
                    )
                self._api.domain.zone.version.set(self._api_key, self._zone_id, version)
                ret = True
            finally:
                if not ret and version is not None:
                    self._api.domain.zone.version.delete(
                        self._api_key, self._zone_id, version
                    )

        LOGGER.debug("delete_record: %s", ret)
        return ret

    @staticmethod
    def _txt_encode(val):
        if not val:
            return None
        return "".join(['"', val.replace("\\", "\\\\").replace('"', '\\"'), '"'])

    @staticmethod
    def _txt_decode(val):
        if not val:
            return None
        if len(val) > 1 and val[0:1] == '"':
            val = val[1:-1].replace('" "', "").replace('\\"', '"').replace("\\\\", "\\")
        return val

    # This exception is for cleaner handling of internal errors
    # within the Gandi provider codebase
    class GandiInternalError(Exception):
        """Internal exception handling class for Gandi management errors"""



"""Integration tests for Henet"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class HenetProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Henet"""

    provider_name = "henet"
    domain = "lexicontest.com"

    def _filter_post_data_parameters(self):
        return ["email", "pass"]

    def _filter_headers(self):
        return ["Authorization", "Cookie"]

    def _filter_query_parameters(self):
        return ["pass"]



"""Module provider for Conoha"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["conoha.io"]


def provider_parser(subparser):
    """Configure provider parser for Conoha"""
    subparser.add_argument(
        "--auth-region", help="specify region. If empty, region `tyo1` will be used."
    )
    subparser.add_argument(
        "--auth-token",
        help=(
            "specify token for authentication. If empty, the username "
            "and password will be used to create a token."
        ),
    )
    subparser.add_argument(
        "--auth-username",
        help="specify api username for authentication. Only used if --auth-token is empty.",
    )
    subparser.add_argument(
        "--auth-password",
        help="specify api user password for authentication. Only used if --auth-token is empty.",
    )
    subparser.add_argument(
        "--auth-tenant-id",
        help="specify tenand id for authentication. Only used if --auth-token is empty.",
    )


class Provider(BaseProvider):
    """Provider class for Conoha"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None

        self.api_endpoint = "https://dns-service.{0}.conoha.io/v1".format(
            self._get_provider_option("region") or "tyo1"
        )
        self.auth_api_endpoint = "https://identity.{0}.conoha.io/v2.0".format(
            self._get_provider_option("region") or "tyo1"
        )
        self.auth_token = None

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def _authenticate(self):
        self.auth_token = self._get_provider_option("auth_token")
        if not self.auth_token:
            if not (
                self._get_provider_option("auth_username")
                and self._get_provider_option("auth_password")
            ):
                raise Exception(
                    "auth_username and auth_password or auth_token must be specified."
                )
            auth_response = self._send_request(
                "POST",
                f"{self.auth_api_endpoint}/tokens",
                {
                    "auth": {
                        "passwordCredentials": {
                            "username": self._get_provider_option("auth_username"),
                            "password": self._get_provider_option("auth_password"),
                        },
                        "tenantId": self._get_provider_option("auth_tenant_id"),
                    }
                },
            )
            self.auth_token = auth_response["access"]["token"]["id"]

        payload = self._get("/domains", {"name": self._fqdn_name(self.domain)})

        if not payload["domains"]:
            raise AuthenticationError("No domain found")
        if len(payload["domains"]) > 1:
            raise AuthenticationError("Too many domains found. This should not happen")

        self.domain_id = payload["domains"][0]["id"]

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        if not rtype:
            raise Exception("rtype must be specified.")
        if not name:
            raise Exception("name must be specified.")
        if not content:
            raise Exception("content must be specified.")
        if not self._get_lexicon_option("priority") and rtype in ("MX", "SRV"):
            raise Exception("priority must be specified.")

        try:
            self._post(
                f"/domains/{self.domain_id}/records",
                self._record_payload(rtype, name, content),
            )
        except requests.exceptions.HTTPError as err:
            # 409 Duplicate Record
            if err.response.status_code != 409:
                raise err

        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/domains/{self.domain_id}/records")
        records = payload["records"]

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._fqdn_name(name)
            ]
        if content:
            records = [record for record in records if record["data"] == content]

        records = [
            {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["data"],
                "id": record["id"],
            }
            for record in records
        ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record. Identifier must be specified.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) != 1:
                raise Exception("Cannot determine record")
            identifier = records[0]["id"]

        self._put(
            f"/domains/{self.domain_id}/records/{identifier}",
            self._record_payload(rtype, name, content),
        )

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        records = self._list_records(rtype, name, content)

        if identifier:
            records = [record for record in records if record["id"] == identifier]

        for record in records:
            self._delete(f"/domains/{self.domain_id}/records/{record['id']}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        return self._send_request(
            action, f"{self.api_endpoint}{url}", data, query_params
        )

    def _send_request(self, action, url, data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            url,
            data=json.dumps(data),
            params=query_params,
            headers={
                "X-Auth-Token": self.auth_token,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if response.headers["content-type"].startswith("application/json"):
            return response.json()
        return response.text

    def _record_name(self, name):
        return f"{name.rstrip('.')}." if name else None

    def _record_payload(self, rtype, name, content):
        priority = self._get_lexicon_option("priority")
        ttl = self._get_lexicon_option("ttl")
        return {
            "name": self._fqdn_name(name) if name else None,
            "type": rtype,
            "data": self._record_name(content)
            if rtype in ("CNAME", "MX", "NS", "SRV")
            else content,
            "priority": int(priority) if priority else None,
            "ttl": int(ttl) if ttl else None,
        }



"""Integration tests for SakuraCloud"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class SakruaCloudProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for SakuraCloud"""

    provider_name = "sakuracloud"
    domain = "example.com"

    def _filter_headers(self):
        return ["Authorization"]

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="record id is not exists")
    def test_provider_when_calling_delete_record_by_identifier_should_remove_record(
        self,
    ):
        return



"""Rackspace provider implementation"""
import json
import logging
import time

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["rackspacecloud.com"]


def _async_request_completed(payload):
    """Looks into an async response payload to see if the requested job has finished."""
    if payload["status"] == "COMPLETED":
        return True
    if payload["status"] == "ERROR":
        return True
    return False


def provider_parser(subparser):
    """Configure provider parser for Rackspace"""
    subparser.add_argument(
        "--auth-account", help="specify account number for authentication"
    )
    subparser.add_argument(
        "--auth-username",
        help="specify username for authentication. Only used if --auth-token is empty.",
    )
    subparser.add_argument(
        "--auth-api-key",
        help="specify api key for authentication. Only used if --auth-token is empty.",
    )
    subparser.add_argument(
        "--auth-token",
        help=(
            "specify token for authentication. "
            "If empty, the username and api key will be used to create a token."
        ),
    )
    subparser.add_argument(
        "--sleep-time",
        type=float,
        default=1,
        help="number of seconds to wait between update requests.",
    )


class Provider(BaseProvider):
    """Provider class for Rackspace"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://dns.api.rackspacecloud.com/v1.0"
        self.auth_api_endpoint = "https://identity.api.rackspacecloud.com/v2.0"
        self._auth_token = None
        self._auth_account = None

    def _get_rackspace_option(self, key):
        private_key = "_" + key
        result = None
        if hasattr(self, private_key):
            result = getattr(self, private_key)
        if result is None:
            result = self._get_provider_option(key)
        return result

    def _authenticate(self):
        self._auth_token = self._get_provider_option("auth_token")
        if not self._auth_token:
            auth_response = self._auth_request(
                "POST",
                "/tokens",
                {
                    "auth": {
                        "RAX-KSKEY:apiKeyCredentials": {
                            "username": self._get_provider_option("auth_username"),
                            "apiKey": self._get_provider_option("auth_api_key"),
                        }
                    }
                },
            )
            self._auth_token = auth_response["access"]["token"]["id"]
            self._auth_account = auth_response["access"]["token"]["tenant"]["id"]

        payload = self._get("/domains", {"name": self.domain})

        if not payload["domains"]:
            raise AuthenticationError("No domain found")
        if len(payload["domains"]) > 1:
            raise AuthenticationError("Too many domains found. This should not happen")

        self.domain_id = payload["domains"][0]["id"]

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        data = {
            "records": [{"type": rtype, "name": self._full_name(name), "data": content}]
        }
        if self._get_lexicon_option("ttl"):
            data["records"][0]["ttl"] = self._get_lexicon_option("ttl")

        try:
            payload = self._post_and_wait(f"/domains/{self.domain_id}/records", data)
        except Exception as error:
            if str(error).startswith("Record is a duplicate of another record"):
                return self._update_record(None, rtype, name, content)
            raise error

        success = len(payload["records"]) > 0
        LOGGER.debug("create_record: %s", success)
        return success

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        params = {"per_page": 100}
        if rtype:
            params["type"] = rtype
        if name:
            params["name"] = self._full_name(name)
        # Sending the data filter to the Rackspace DNS API results in a 503 error
        # if content:
        #     params['data'] = content

        payload = self._get(f"/domains/{self.domain_id}/records", params)

        records = list(payload["records"])
        if content:
            records = [record for record in records if record["data"] == content]
        records = [
            {
                "type": record["type"],
                "name": record["name"],
                "ttl": record["ttl"],
                "content": record["data"],
                "id": record["id"],
            }
            for record in records
        ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._full_name(name)
        if content:
            data["data"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        if identifier is None:
            records = self._list_records(rtype, name)
            if not records:
                raise Exception("Unable to find record to modify: " + name)
            identifier = records[0]["id"]

        self._put_and_wait(f"/domains/{self.domain_id}/records/{identifier}", data)

        # If it didn't raise from the http status code, then we're good
        LOGGER.debug("update_record: %s", identifier)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete_and_wait(f"/domains/{self.domain_id}/records/{record_id}")

        # If it didn't raise from the http status code, then we're good
        success = True
        LOGGER.debug("delete_record: %s", success)
        return success

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        LOGGER.debug(
            "request tenant ID: %s", self._get_rackspace_option("auth_account")
        )
        full_url = (
            f"{self.api_endpoint}/{self._get_rackspace_option('auth_account')}{url}"
        )
        # For data= argument, use None value for GET requests to comply with Rackspace API
        response = requests.request(
            action,
            full_url,
            params=query_params,
            data=json.dumps(data) if action != "GET" else None,
            headers={
                "X-Auth-Token": self._get_rackspace_option("auth_token"),
                "Content-Type": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    # Non-GET requests to the Rackspace CloudDNS API are asynchronous
    def _request_and_wait(self, action="POST", url="/", data=None, query_params=None):
        result = self._request(action, url, data, query_params)

        sleep_time = self._get_rackspace_option("sleep_time") or "1"
        sleep_time = float(sleep_time)

        while not _async_request_completed(result):
            if sleep_time:
                time.sleep(sleep_time)
            result = self._update_response(result)

        if result["status"] == "ERROR":
            raise Exception(result["error"]["details"])

        if "response" in result:
            return result["response"]
        return None

    def _post_and_wait(self, url="/", data=None, query_params=None):
        return self._request_and_wait("POST", url, data, query_params)

    def _put_and_wait(self, url="/", data=None, query_params=None):
        return self._request_and_wait("PUT", url, data, query_params)

    def _delete_and_wait(self, url="/", data=None, query_params=None):
        return self._request_and_wait("DELETE", url, data, query_params)

    def _update_response(self, payload):
        response = requests.request(
            "GET",
            payload["callbackUrl"],
            params={"showDetails": "true"},
            data={},
            headers={
                "X-Auth-Token": self._get_rackspace_option("auth_token"),
                "Content-Type": "application/json",
            },
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _auth_request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}

        response = requests.request(
            action,
            self.auth_api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Aliyun"""
import base64
import datetime
import hmac
import logging
import random
import sys
import time
from hashlib import sha1

import requests
from six.moves import urllib

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["hichina.com"]

ALIYUN_DNS_API_ENDPOINT = "https://alidns.aliyuncs.com"


def provider_parser(subparser):
    """Module provider for Aliyun"""
    subparser.description = """
        Aliyun Provider requires an access key id and access secret with full rights on dns.
        Better to use RAM on Aliyun cloud to create a specified user for the dns operation.
        The referrence for Aliyun DNS production:
        https://help.aliyun.com/product/29697.html"""
    subparser.add_argument(
        "--auth-key-id", help="specify access key id for authentication"
    )
    subparser.add_argument(
        "--auth-secret", help="specify access secret for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Aliyun"""

    def _authenticate(self):
        response = self._request_aliyun("DescribeDomainInfo")

        if "DomainId" not in response:
            raise AuthenticationError(
                f"failed to fetch basic domain info for {self.domain}"
            )

        self.domain_id = response["DomainId"]

        return self

    def _create_record(self, rtype, name, content):
        if not self._list_records(rtype, name, content):
            query_params = {
                "Value": content,
                "Type": rtype,
                "RR": self._relative_name(name),
                "TTL": self._get_lexicon_option("ttl"),
            }
            self._request_aliyun("AddDomainRecord", query_params=query_params)

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        query_params = {}

        if rtype:
            query_params["TypeKeyWord"] = rtype

        if name:
            query_params["RRKeyWord"] = self._relative_name(name)

        if content:
            query_params["ValueKeyWord"] = content

        response = self._request_aliyun(
            "DescribeDomainRecords", query_params=query_params
        )

        resource_list = response["DomainRecords"]["Record"]

        processed_records = []
        for resource in resource_list:
            processed_records.append(
                {
                    "id": resource["RecordId"],
                    "type": resource["Type"],
                    "name": self._full_name(resource["RR"]),
                    "ttl": resource["TTL"],
                    "content": resource["Value"],
                }
            )
        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        resources = self._list_records(rtype, name, None)

        for record in resources:
            if (
                rtype == record["type"]
                and (self._relative_name(name) == self._relative_name(record["name"]))
                and (content == record["content"])
            ):
                return True

        if not identifier:
            record = resources[0] if resources else None
            identifier = record["id"] if record else None

        if not identifier:
            raise ValueError(f"updating {identifier} identifier not exists")

        if len(resources) > 1:
            LOGGER.warning(
                """There's more than one records match the given critiaria,
             only the first one would be updated"""
            )

        LOGGER.debug("update_record: %s", identifier)

        query_params = {"RecordId": identifier}

        if rtype:
            query_params["Type"] = rtype

        if name:
            query_params["RR"] = self._relative_name(name)

        if content:
            query_params["Value"] = content

        query_params["TTL"] = self._get_lexicon_option("ttl")
        self._request_aliyun("UpdateDomainRecord", query_params=query_params)

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self._list_records(rtype, name, content)
            delete_resource_id = [resource["id"] for resource in resources]
        else:
            delete_resource_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_resource_id)

        for resource_id in delete_resource_id:
            self._request_aliyun(
                "DeleteDomainRecord", query_params={"RecordId": resource_id}
            )

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        response = requests.request("GET", ALIYUN_DNS_API_ENDPOINT, params=query_params)
        response.raise_for_status()

        try:
            return response.json()
        except ValueError as invalid_json_ve:
            LOGGER.error(
                "aliyun dns api responsed with invalid json content, %s", response.text
            )
            raise invalid_json_ve

    def _request_aliyun(self, action, query_params=None):
        if query_params is None:
            query_params = {}

        query_params.update(self._build_default_query_params(action))
        query_params.update(self._build_signature_parameters())

        query_params.update(
            {"Signature": self._calculate_signature("GET", query_params)}
        )

        return self._request(url=ALIYUN_DNS_API_ENDPOINT, query_params=query_params)

    def _calculate_signature(self, http_method, query_params):
        access_secret = self._get_provider_option("auth_secret")

        if not access_secret:
            raise ValueError(
                "auth-secret (access secret) is not specified, did you forget that?"
            )

        sign_secret = access_secret + "&"

        query_list = list(query_params.items())
        query_list.sort(key=lambda t: t[0])

        canonicalized_query_string = urllib.parse.urlencode(query_list)

        string_to_sign = "&".join(
            [
                http_method,
                urllib.parse.quote_plus("/"),
                urllib.parse.quote_plus(canonicalized_query_string),
            ]
        )

        if sys.version_info.major > 2:
            sign_secret_bytes = bytes(sign_secret, "utf-8")
            string_to_sign_bytes = bytes(string_to_sign, "utf-8")
            sign = hmac.new(sign_secret_bytes, string_to_sign_bytes, sha1)
            signature = base64.b64encode(sign.digest()).decode()
        else:
            sign = hmac.new(sign_secret, string_to_sign, sha1)
            signature = sign.digest().encode("base64").rstrip("\n")

        return signature

    def _build_signature_parameters(self):
        access_key_id = self._get_provider_option("auth_key_id")

        if not access_key_id:
            raise ValueError(
                "auth-key-id (access key id) is not specified, did you forget that?"
            )

        signature_nonce = str(int(time.time())) + str(random.randint(1000, 9999))

        return {
            "SignatureMethod": "HMAC-SHA1",
            "SignatureVersion": "1.0",
            "SignatureNonce": signature_nonce,
            "Timestamp": datetime.datetime.utcnow().replace(microsecond=0).isoformat()
            + "Z",
            "AccessKeyId": access_key_id,
        }

    def _build_default_query_params(self, action):
        return {
            "Action": action,
            "DomainName": self.domain,
            "Format": "json",
            "Version": "2015-01-09",
        }
